/*****************************************************************************/
/*                                                                           */
/* Author: Michael C. Nechyba                                                */
/* Last modified: 2/25/2001                                                  */
/*                                                                           */
/* File: "ellipse.c"                                                         */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef INCLUDE_JAVA
#include "GatorCameraClient.h"
#endif

#include "acticast_include.h"

/****************************** Subroutines **********************************/


int copy_ellipse(ELLIPSE *ein, ELLIPSE *eout)
{
  eout->cov[0][0] = ein->cov[0][0];
  eout->cov[0][1] = ein->cov[0][1];
  eout->cov[1][0] = ein->cov[1][0];
  eout->cov[1][1] = ein->cov[1][1];

  eout->mean[0] = ein->mean[0];
  eout->mean[1] = ein->mean[1];

  eout->major_axis[0] = ein->major_axis[0];
  eout->major_axis[1] = ein->major_axis[1];

  eout->ev[0] = ein->ev[0];
  eout->ev[1] = ein->ev[1];
  eout->theta = ein->theta;
  eout->a = ein->a;
  eout->b = ein->b;

  return(1);
}

/*****************************************************************************/

int initialize_ellipse(ELLIPSE *e)
{
  e->cov[0][0] = 0.;
  e->cov[0][1] = 0.;
  e->cov[1][0] = 0.;
  e->cov[1][1] = 0.;

  e->mean[0] = 0.;
  e->mean[1] = 0.;

  e->major_axis[0] = 0.;
  e->major_axis[1] = 0.;

  e->ev[0] = 0.;
  e->ev[1] = 0.;
  e->theta = 0.;
  e->a = 0.;
  e->b = 0.;

  e->xmin = 1000;
  e->xmax = -1000;
  e->ymin = 1000;
  e->ymax = -1000;
  
  return(1);
}

/*****************************************************************************/

int compute_ellipse_bounds(ELLIPSE *e, double **x, int n, int transpose)
{
  int j; 

  if(transpose == 0) {
    for(j = 0; j < n; j++) {
      e->xmin = min_int(e->xmin, x[j][0]);
      e->xmax = max_int(e->xmax, x[j][0]);
      
      e->ymin = min_int(e->ymin, x[j][1]);
      e->ymax = max_int(e->ymax, x[j][1]);
    }      
  } else {
    for(j = 0; j < n; j++) {
      e->xmin = min_int(e->xmin, x[0][j]);
      e->xmax = max_int(e->xmax, x[0][j]);
      
      e->ymin = min_int(e->ymin, x[1][j]);
      e->ymax = max_int(e->ymax, x[1][j]);
    }
  }

  return(1);
}

/*****************************************************************************/

int mean_matrix(double *v, double **x, int n, int m, int transpose)
{
  int i, j;

  for(i = 0; i < m; i++)
    v[i] = 0.;

  for(i = 0; i < m; i++) {
    if(transpose == 0) {
      for(j = 0; j < n; j++)
	v[i] += x[j][i];
    } else {
      for(j = 0; j < n; j++)
	v[i] += x[i][j];
    }
  }

  for(i = 0; i < m; i++)
    v[i] /= ((double) n);

  return(1);
}

/*****************************************************************************/

int covariance_matrix2x2(double cov[2][2], double **x, int n, double *mean,
			 int transpose)
{
  int i, j, k;

  for(j = 0; j < 2; j++)
    for(k = 0; k < 2; k++) {
      if(j <= k) {
	cov[j][k] = 0.;
	
	if(transpose == 0) {
	  for(i = 0; i < n; i++) 
	    cov[j][k] += (x[i][j] - mean[j])*(x[i][k] - mean[k]);
	} else {
	  for(i = 0; i < n; i++) 
	    cov[j][k] += (x[j][i] - mean[j])*(x[k][i] - mean[k]);
	}

	cov[j][k] /= ((double) (n-1));
      } else {
	cov[j][k] = cov[k][j];
      }
    }

  return(1);
}

/*****************************************************************************/

int eigenvalues2x2(double *ev, double m[2][2])
{
  double a, b, c, d, rad;
  double eps = 0.000000001;

  a = m[0][0];
  b = m[0][1];
  c = m[1][0];
  d = m[1][1];

  rad = a*a + 4.*b*c - 2.*a*d + d*d;

  if(rad < -eps) {
    fprintf(stderr, "Warning: eigenvalue is complex. ");
    fprintf(stderr, "Unimplemented and buggin' out (%.15f).\n", rad);
    fprintf(stderr, "cov = [%.15f %.15f]\n      [%.15f %.15f]\n", 
	    m[0][0], m[0][1], m[1][0], m[1][1]);
    exit(-1);
  }

  rad = fabs(rad);

  ev[0] = .5*(a + d - sqrt(rad));
  ev[1] = .5*(a + d + sqrt(rad));

  return(1);
}

/*****************************************************************************/

int major_eigenvector2x2(double *ev, double m[2][2])
{
  double evalues[2], norm;

  eigenvalues2x2(&evalues[0], m);

  if(is_zero2(m[0][1], 0.000001)) {
    m[0][1] = 0.000001;
  }

  ev[1] = m[0][0] - evalues[1];
  ev[0] = -m[0][1];

  norm = sqrt(ev[1]*ev[1] + ev[0]*ev[0]);

  ev[0] /= norm;
  ev[1] /= norm;

  if(ev[1] < 0.) {
    ev[0] = -ev[0];
    ev[1] = -ev[1];
  }

  return(1);
}

/*****************************************************************************/

double ellipsoid_angle(double m[2][2])
{
  double evec[2];
  major_eigenvector2x2(&evec[0], m);

  return(atan2(evec[1], evec[0]));
}

/*****************************************************************************/

/* 
   Given a percentage value 0 < x < 1, this function returns 
   the length of the axis (for a standard deviation of 1) for
   which a 2-d Gaussian encompasses x of the total mass under
   that Gaussian.

   Values were derived through Mathematica.
*/

double ellipsoid_distance(double x)
{
  int index;
  double lut[100] = 
  {0.000000000000000,  0.141776837695735,  0.201010981379224,  
   0.246816561375887,  0.28573412298938,  0.320291412271858,  
   0.351782329624691,  0.380974258539433,  0.408366523943996,  
   0.434305605469792,  0.459043605026421,  0.482770786721714,  
   0.505634989908501,  0.527753858031389,  0.549222886876692,  
   0.570120916118283,  0.590513991612016,  0.6104581528516,  
   0.630001490036076,  0.649185691949002,  0.668047230836577,  
   0.686618283358475,  0.704927456265536,  0.723000365331038,  
   0.740860102450875,  0.758527616440932,  0.776022026470798,  
   0.793360882372833,  0.810560382663792,  0.827635558620793,  
   0.844600430900592,  0.861468143799679,  0.878251081197154,  
   0.894960967413803,  0.911608955596275,  0.928205705748951,  
   0.944761454154878,  0.961286075626355,  0.977789139787306,  
   0.994279962399706,  1.01076765259479,  1.02726115674873,  
   1.04376929964592,  1.06030082349637,  1.07686442531355,  
   1.09346879311265,  1.11012264135438,  1.12683474603508,  
   1.14361397980845,  1.16046934751743,  1.17741002251547,  
   1.19444538416578,  1.21158505692353,  1.22883895143182,  
   1.24621730809598,  1.26373074364579,  1.28139030125082,  
   1.29920750482325,  1.31719441822741,  1.33536371021814,  
   1.35372872605567,  1.37230356689651,  1.39110317824503,  
   1.41014344897522,  1.42944132270757,  1.44901492366275,  
   1.46888369952963,  1.48906858439873,  1.50959218545166,  
   1.53047899789768,  1.55175565365552,  1.57345121055698,  
   1.59559749047991,  1.61822947691838,  1.64138578522333,  
   1.6651092223154,  1.68944745738963,  1.71445383143376,  
   1.74018834189278,  1.76671885044829,  1.7941225779941,  
   1.82248797352501,  1.85191707594694,  1.88252853467451,  
   1.91446152416198,  1.94788089209062,  1.98298404248387,  
   2.02001031112544,  2.05925400871291,  2.10108301272925,  
   2.14596602628935,  2.19451389088876,  2.24754472449749,  
   2.30619168194354,  2.37209220594817,  2.44774683068082,  
   2.53727248235904,  2.64822880330231,  2.79714962253654,  
   3.03485425877029};
  
  index = (int) (x*100. + 0.5);
  index = limit_int(index, 1, 99);
  
  return(lut[index]);
}

/*****************************************************************************/

int compute_ellipse(ELLIPSE *ell, double **x, int n, double ci,
		    int transpose)
{
  compute_ellipse_bounds(ell, x, n, transpose);
  mean_matrix(&ell->mean[0], x, n, 2, transpose);
  covariance_matrix2x2(ell->cov, x, n, &ell->mean[0], transpose);
  compute_ellipse_from_cov(ell, ci);
  return(1);
}

/*****************************************************************************/

int compute_ellipse_from_cov(ELLIPSE *ell, double ci)
{
  double l;

  eigenvalues2x2(&ell->ev[0], ell->cov);
  major_eigenvector2x2(&ell->major_axis[0], ell->cov);
  ell->theta = ellipsoid_angle(ell->cov);

  l = ellipsoid_distance(ci);
  ell->a = sqrt(ell->ev[1])*l;
  ell->b = sqrt(ell->ev[0])*l;

  return(1);
}

/*****************************************************************************/

int compute_ellipse_from_online(ELLIPSE *ell, int n)
{
  double nd, xd, yd;

  nd = ((double) n);

  ell->mean[0] /= nd;
  ell->mean[1] /= nd;

  xd = ell->mean[0];
  yd = ell->mean[1];
  
  ell->cov[0][0] = nd/(nd - 1.)*(ell->cov[0][0]/nd - xd*xd);
  ell->cov[0][1] = nd/(nd - 1.)*(ell->cov[0][1]/nd - xd*yd);
  ell->cov[1][0] = ell->cov[0][1];
  ell->cov[1][1] = nd/(nd - 1.)*(ell->cov[1][1]/nd - yd*yd);

  compute_ellipse_from_cov(ell, CONFIDENCE_INTERVAL);

  return(1);
}

/*****************************************************************************/

int display_ellipse(FILE *fp, ELLIPSE *ell)
{
  fprintf(fp, "ELLIPSE: (xmin, ymin) = ( %d , %d ); (xmax, ymax) = ( %d , %d ).\n", 
	  ell->xmin, ell->ymin, ell->xmax, ell->ymax);

  fprintf(fp, "ELLIPSE: Ellipse mean (xc, yc) = ( %lf ,  %lf ).\n", 
	  ell->mean[0], ell->mean[1]);

  fprintf(fp, "ELLIPSE: Major axis = (%lf, %lf)\n",
	  ell->major_axis[0], ell->major_axis[1]);

  fprintf(fp, "ELLIPSE: Covariance matrix:\n");
  fprintf(fp, "ELLIPSE:\t%lf\t%lf\nELLIPSE:\t%lf\t%lf\n",
	  ell->cov[0][0], ell->cov[0][1], ell->cov[1][0], ell->cov[1][1]);

  fprintf(fp, "ELLIPSE: Eigenvalues = %lf, %lf.\n", 
	  ell->ev[0], ell->ev[1]);

  fprintf(fp, "ELLIPSE: Normalized axes length = %lf, %lf.\n", 
	  ell->a, ell->b);

  fprintf(fp, "ELLIPSE: Ellipse angle = %lf (%lf deg).\n", 
	  ell->theta, ell->theta * 180./M_PI);

  return(1);
}

/*****************************************************************************/

int selective_ellipse(ELLIPSE *e, double **x, 
		      int n, int transpose, int *type, int pick)
{
  int i, j, ns = 0;

  e->xmin = 1000;
  e->xmax = -1000;
  e->ymin = 1000;
  e->ymax = -1000;

  e->cov[0][0] = 0.;
  e->cov[0][1] = 0.;
  e->cov[1][0] = 0.;
  e->cov[1][1] = 0.;

  e->mean[0] = 0.;
  e->mean[1] = 0.;

  if(transpose == 0) {
    for(j = 0; j < n; j++) {
      if(type[j] == pick) {
	e->xmin = min_int(e->xmin, x[j][0]);
	e->xmax = max_int(e->xmax, x[j][0]);

	e->ymin = min_int(e->ymin, x[j][1]);
	e->ymax = max_int(e->ymax, x[j][1]);


	e->mean[0] += x[j][0];
	e->mean[1] += x[j][1];
	e->cov[0][0] += x[j][0]*x[j][0];
	e->cov[0][1] += x[j][1]*x[j][0];
	e->cov[1][1] += x[j][1]*x[j][1];
	ns++;
      }
    }
    e->cov[1][0] = e->cov[0][1];
  } else {
    for(j = 0; j < n; j++) {
      if(type[j] == pick) {
	e->xmin = min_int(e->xmin, x[0][j]);
	e->xmax = max_int(e->xmax, x[0][j]);

	e->ymin = min_int(e->ymin, x[1][j]);
	e->ymax = max_int(e->ymax, x[1][j]);

	e->mean[0] += x[0][j];
	e->mean[1] += x[1][j];
	e->cov[0][0] += x[0][j]*x[0][j];
	e->cov[0][1] += x[1][j]*x[0][j];
	e->cov[1][1] += x[1][j]*x[1][j];
	ns++;
      }
    }
    e->cov[1][0] = e->cov[0][1];
  }

  compute_ellipse_from_online(e, ns);

  return(ns);
}

/*****************************************************************************/

double ellipse_log_likelihood(double *xy, ELLIPSE *e) 
{
  double x, y, xd, yd;
  double m0, m1, c00, c01, c11, detc;
  double lprob, l4;

  x = xy[0];
  y = xy[1];
  m0 = e->mean[0];
  m1 = e->mean[1];
  c00 = e->cov[0][0];
  c01 = e->cov[0][1];
  c11 = e->cov[1][1];


  detc = c11*c00 - c01*c01;
  xd = (x-m0); yd = (y-m1);

  lprob = -0.5*(yd*(-(c01*xd) + c00*yd) + xd*(c11*xd - c01*yd))/detc;
  lprob -= 0.5*log(detc);

  // doesn't add -ln(2.) - ln(pi)

#ifndef SUPPRESS_ALL_OUTPUT
#ifdef LEVEL_TWO_DIAGNOSTIC
  fprintf(stderr, 
	  "E:\n{x->%.15f, y->%.15f, m0->%.15f, m1->%.15f, ",
	  x, y, m0, m1);
  fprintf(stderr, 
	  "c00->%.15f,c01-> %.15f,c11->%.15f}\nlprob = %.15f\n",
	  c00, c01, c11, lprob);
#endif
#endif
  return(lprob);
}

/*****************************************************************************/

int compute_position_class_one(double *xy, ELLIPSE *e, int n_e) 
{
  int ind, i;
  double tst, p = -10000000.;

  for(i = 0; i < n_e; i++) {
    tst = ellipse_log_likelihood(xy, &e[i]);
    //fprintf(stderr, "%d --> %.15f\n", i, tst);
    if(tst > p) {
      p = tst;
      ind = i;
    }
  }

  return(ind);
}

/*****************************************************************************/

int compute_position_class(BLOB2 *b, int *w, int n_w)
{
  int i, ind;
  double xy[2];
  ELLIPSE e[N_OBJECTS_MAX];

  for(i = 0; i < n_w; i++) {
    selective_ellipse(&e[i], b->xy_t, b->n, 1, b->type, w[i]);
    b->tags[i] = w[i];
    //display_ellipse(stderr, &e[i]);
  }

  b->n_tags = n_w;

  for(i = 0; i < b->n; i++) {
    xy[0] = b->xy_t[0][i];
    xy[1] = b->xy_t[1][i];
    b->type[i] = w[compute_position_class_one(&xy[0], &e[0], n_w)];
  }

  return(1);
}

/*****************************************************************************/
