/*****************************************************************************/
/*                                                                           */
/* Author: Michael C. Nechyba                                                */
/* Last modified: 2/25/2001                                                  */
/*                                                                           */
/* File: "geometry.c"                                                        */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef INCLUDE_JAVA
#include "GatorCameraClient.h"
#endif

#include "acticast_include.h"

/****************************** Subroutines **********************************/

int rotation_euler(double *vout, double *vin, 
		   double alpha, double beta, double gamma) 
{
  int i;
  double r[3][3], ca, sa, cb, sb, cg, sg;
  
  ca = cos(alpha);
  sa = sin(alpha);
  cb = cos(beta);
  sb = sin(beta);
  cg = cos(gamma);
  sg = sin(gamma);
  
  r[0][0] = ca*cb;
  r[0][1] = -(cg*sa) + ca*sb*sg;
  r[0][2] = ca*cg*sb + sa*sg;
  r[1][0] = cb*sa;
  r[1][1] = ca*cg + sa*sb*sg;
  r[1][2] = cg*sa*sb - ca*sg;
  r[2][0] = -sb;
  r[2][1] = cb*sg;
  r[2][2] = cb*cg;
  
  for(i = 0; i < 3; i++) {
    vout[i] = r[i][0]*vin[0] + r[i][1]*vin[1] + r[i][2]*vin[2];
  }
  
  return(1);
}

/*****************************************************************************/

int rotation_fixed(double *vout, double *vin, 
		   double alpha, double beta, double gamma) 
{
  int i;
  double r[3][3], ca, sa, cb, sb, cg, sg;
  
  ca = cos(alpha);
  sa = sin(alpha);
  cb = cos(beta);
  sb = sin(beta);
  cg = cos(gamma);
  sg = sin(gamma);
  
  r[0][0] = ca*cb;
  r[0][1] = ca*sb*sg - sa*cg;
  r[0][2] = ca*sb*cg + sa*sg;
  r[1][0] = sa*cb;
  r[1][1] = sa*sb*sg + ca*cg;
  r[1][2] = sa*sb*cg - ca*sg;
  r[2][0] = -sb;
  r[2][1] = cb*sg;
  r[2][2] = cb*cg;
  
  for(i = 0; i < 3; i++) {
    vout[i] = r[i][0]*vin[0] + r[i][1]*vin[1] + r[i][2]*vin[2];
  }
  
  return(1);
}

/*****************************************************************************/

int rotation_euler_trans(double *vout, double *vin, 
			 double alpha, double beta, double gamma) 
{
  int i;
  double r[3][3], ca, sa, cb, sb, cg, sg;
  
  ca = cos(alpha);
  sa = sin(alpha);
  cb = cos(beta);
  sb = sin(beta);
  cg = cos(gamma);
  sg = sin(gamma);
  
  r[0][0] = ca*cb;
  r[1][0] = -(cg*sa) + ca*sb*sg;
  r[2][0] = ca*cg*sb + sa*sg;
  r[0][1] = cb*sa;
  r[1][1] = ca*cg + sa*sb*sg;
  r[2][1] = cg*sa*sb - ca*sg;
  r[0][2] = -sb;
  r[1][2] = cb*sg;
  r[2][2] = cb*cg;
  
  for(i = 0; i < 3; i++) {
    vout[i] = r[i][0]*vin[0] + r[i][1]*vin[1] + r[i][2]*vin[2];
  }
  
  return(1);
}

/*****************************************************************************/

int translation(double *vout, double *vin, double x0, double y0, double z0) 
{
  vout[0] = vin[0] + x0;
  vout[1] = vin[1] + y0;
  vout[2] = vin[2] + z0;	
  
  return(1);
}

/*****************************************************************************/

int forward_transformation(double *vout, double *vin, 
			   double alpha, double beta, double gamma, 
			   double x0, double y0, double z0) 
{	
  double vtmp[3];
  
  rotation_euler(&vtmp[0], vin, alpha, beta, gamma);
  translation(vout, &vtmp[0], x0, y0, z0);
  
  return(1);
}

/*****************************************************************************/

int inverse_transformation(double *vout, double *vin, 
			   double alpha, double beta, double gamma, 
			   double x0, double y0, double z0) 
{	
  double vtmp[3];
  
  translation(&vtmp[0], vin, -x0, -y0, -z0);
  rotation_euler_trans(vout, &vtmp[0], alpha, beta, gamma);
  
  return(1);
}

/*****************************************************************************/

double dot_local(double *v1, double *v2)
{
  double out;
  
  out = v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2];
  
  return(out);
}

/*****************************************************************************/

int cross(double *vout, double *v1, double *v2)
{
  vout[0] = v1[1]*v2[2] - v2[1]*v1[2];
  vout[1] = v1[2]*v2[0] - v2[2]*v1[0];
  vout[2] = v1[0]*v2[1] - v2[0]*v1[1];
  
  return(1);
}

/*****************************************************************************/

int normalize_local(double *vout, double *vin)
{
  double sum;
  
  sum = sqrt(dot_local(vin, vin));
  
  vout[0] = vin[0]/sum;
  vout[1] = vin[1]/sum;
  vout[2] = vin[2]/sum;
  
  return(1);
}
/*****************************************************************************/

int plane_equation(double *v, TRAPEZOID *zoid)
{
  double *c1, *c2, *c3;
  
  c1 = zoid->c1;
  c2 = zoid->c2;
  c3 = zoid->c3;
  
  v[0] = -(c1[2]*c2[1]) + c1[1]*c2[2] + c1[2]*c3[1] - c2[2]*c3[1] - 
    c1[1]*c3[2] + c2[1]*c3[2];
  v[1] = c1[2]*c2[0] - c1[0]*c2[2] - c1[2]*c3[0] + c2[2]*c3[0] + 
    c1[0]*c3[2] - c2[0]*c3[2];
  v[2] = -(c1[1]*c2[0]) + c1[0]*c2[1] + c1[1]*c3[0] - c2[1]*c3[0] - 
    c1[0]*c3[1] + c2[0]*c3[1];
  v[3] = -(c1[2]*c2[1]*c3[0]) + c1[1]*c2[2]*c3[0] + c1[2]*c2[0]*c3[1] - 
    c1[0]*c2[2]*c3[1] - c1[1]*c2[0]*c3[2] + c1[0]*c2[1]*c3[2];
	
  return(1);
}

/*****************************************************************************/

int rectangles_overlap(BOUNDING_BOX *b1, BOUNDING_BOX *b2)
{
  if(check_overlap_corner(b1->xmin, b1->ymin, b2) ||
     check_overlap_corner(b1->xmax, b1->ymin, b2) ||
     check_overlap_corner(b1->xmin, b1->ymax, b2) ||
     check_overlap_corner(b1->xmax, b1->ymax, b2))
    return(1);
  else if (check_overlap_corner(b2->xmin, b2->ymin, b1) ||
	   check_overlap_corner(b2->xmax, b2->ymin, b1) ||
	   check_overlap_corner(b2->xmin, b2->ymax, b1) ||
	   check_overlap_corner(b2->xmax, b2->ymax, b1))
    return(1);
  else if (b1->xmin > b2->xmin && b1->xmax < b2->xmax && 
	   b1->ymin < b2->ymin && b1->ymax > b2->ymax)
    return(1);
  else if (b1->xmin < b2->xmin && b1->xmax > b2->xmax && 
	   b1->ymin > b2->ymin && b1->ymax < b2->ymax)
    return(1);
  else
    return(0);
}

/*****************************************************************************/

int check_overlap_corner(int x, int y, BOUNDING_BOX *b)
{
  if(x <= b->xmax && x >= b->xmin && y <= b->ymax && y >= b->ymin)
    return(1);
  else
    return(0);
}

/*****************************************************************************/

int overlap_merge(BOUNDING_BOX *b1, BOUNDING_BOX *b2) 
{
  BOUNDING_BOX btmp;

  btmp.xmin = min_int(b1->xmin, b2->xmin);
  btmp.xmax = max_int(b1->xmax, b2->xmax);
  btmp.ymin = min_int(b1->ymin, b2->ymin);
  btmp.ymax = max_int(b1->ymax, b2->ymax);

  b1->xmin = btmp.xmin;
  b1->xmax = btmp.xmax;
  b1->ymin = btmp.ymin;
  b1->ymax = btmp.ymax;

  b2->xmin = btmp.xmin;
  b2->xmax = btmp.xmax;
  b2->ymin = btmp.ymin;
  b2->ymax = btmp.ymax;

  return(1);
}

/*****************************************************************************/

int copy_rotation_matrix(RotationMatrix *in, RotationMatrix *out)
{
  out->r11 = in->r11;
  out->r12 = in->r12;
  out->r13 = in->r13;

  out->r21 = in->r21;
  out->r22 = in->r22;
  out->r23 = in->r23;

  out->r31 = in->r31;
  out->r32 = in->r32;
  out->r33 = in->r33;

  return(1);
}

/*****************************************************************************/

int identity_rotation_matrix(RotationMatrix *m)
{
  m->r11 = 1.;
  m->r22 = 1.;
  m->r33 = 1.;

  m->r12 = 0.;
  m->r13 = 0.;

  m->r21 = 0.;
  m->r23 = 0.;

  m->r31 = 0.;
  m->r32 = 0.;

  return(1);
}

/*****************************************************************************/

int rotation_matrix_x(RotationMatrix *m, double alpha)
{
  double ca, sa;
  RotationMatrix m2;

  ca = cos(alpha);
  sa = sin(alpha);

  m2.r11 = m->r11;
  m2.r12 = m->r12;
  m2.r13 = m->r13;

  m2.r21 = (m->r21)*ca - (m->r31)*sa;
  m2.r22 = (m->r22)*ca - (m->r32)*sa;
  m2.r23 = (m->r23)*ca - (m->r33)*sa;

  m2.r31 = (m->r31)*ca + (m->r21)*sa;
  m2.r32 = (m->r32)*ca + (m->r22)*sa;
  m2.r33 = (m->r33)*ca + (m->r23)*sa;

  copy_rotation_matrix(&m2, m);

  return(1);
}

/*****************************************************************************/

int rotation_matrix_y(RotationMatrix *m, double alpha)
{
  double ca, sa;
  RotationMatrix m2;

  ca = cos(alpha);
  sa = sin(alpha);

  m2.r11 = (m->r11)*ca + (m->r31)*sa;
  m2.r12 = (m->r12)*ca + (m->r32)*sa;
  m2.r13 = (m->r13)*ca + (m->r33)*sa;

  m2.r21 = m->r21;
  m2.r22 = m->r22;
  m2.r23 = m->r23;

  m2.r31 = (m->r31)*ca - (m->r11)*sa;
  m2.r32 = (m->r32)*ca - (m->r12)*sa;
  m2.r33 = (m->r33)*ca - (m->r13)*sa;


  copy_rotation_matrix(&m2, m);

  return(1);
}

/*****************************************************************************/

int rotation_matrix_z(RotationMatrix *m, double alpha)
{
  double ca, sa;
  RotationMatrix m2;

  ca = cos(alpha);
  sa = sin(alpha);

  m2.r11 = (m->r11)*ca - (m->r21)*sa;
  m2.r12 = (m->r12)*ca - (m->r22)*sa;
  m2.r13 = (m->r13)*ca - (m->r23)*sa;

  m2.r21 = (m->r21)*ca + (m->r11)*sa;
  m2.r22 = (m->r22)*ca + (m->r12)*sa;
  m2.r23 = (m->r23)*ca + (m->r13)*sa;

  m2.r31 = m->r31;
  m2.r32 = m->r32;
  m2.r33 = m->r33;

  copy_rotation_matrix(&m2, m);

  return(1);
}

/*****************************************************************************/

int display_rotation_matrix(FILE *fp, RotationMatrix *m)
{
  fprintf(fp, "%lf %lf %lf\n", m->r11, m->r12, m->r13);
  fprintf(fp, "%lf %lf %lf\n", m->r21, m->r22, m->r23);
  fprintf(fp, "%lf %lf %lf\n", m->r31, m->r32, m->r33);

  return(1);
}

/*****************************************************************************/

int display_rotation_constraints(FILE *fp, RotationMatrix *m)
{
  double a1,a2,a3,a4,a5,a6;

  a1 = m->r11*m->r21 + m->r12*m->r22 + m->r13*m->r23;
  a2 = m->r31*m->r21 + m->r32*m->r22 + m->r33*m->r23;
  a3 = m->r31*m->r11 + m->r32*m->r12 + m->r33*m->r13;

  a4 = m->r11*m->r11 + m->r12*m->r12 + m->r13*m->r13;
  a5 = m->r21*m->r21 + m->r22*m->r22 + m->r23*m->r23;
  a6 = m->r31*m->r31 + m->r32*m->r32 + m->r33*m->r33;

  fprintf(fp, "normal: %lf %lf %lf\n", a4, a5, a6);
  fprintf(fp, "orthog: %lf %lf %lf\n", a1, a2, a3);

  return(1);
}

/*****************************************************************************/

int rXt(TranslationVector *out, RotationMatrix *m, TranslationVector *in)
{
  out->x = (m->r11)*(in->x) + (m->r12)*(in->y) + (m->r13)*(in->z);
  out->y = (m->r21)*(in->x) + (m->r22)*(in->y) + (m->r23)*(in->z);
  out->z = (m->r31)*(in->x) + (m->r32)*(in->y) + (m->r33)*(in->z);

  return(1);
}

/*****************************************************************************/

int rotation_matrix_angle_axis(RotationMatrix *m, 
			       TranslationVector *k, double theta)
{
  double ct, st, vt, kx, ky, kz;
  RotationMatrix m1, m2;

  ct = cos(theta);
  st = sin(theta);
  vt = 1. - ct;

  kx = k->x;
  ky = k->y;
  kz = k->z;

  m2.r11 = kx*kx*vt + ct;
  m2.r12 = kx*ky*vt - kz*st;
  m2.r13 = kx*kz*vt + ky*st;

  m2.r21 = kx*ky*vt + kz*st;
  m2.r22 = ky*ky*vt + ct;
  m2.r23 = ky*kz*vt - kx*st;

  m2.r31 = kx*kz*vt - ky*st;
  m2.r32 = ky*kz*vt + kx*st;
  m2.r33 = kz*kz*vt + ct;


  copy_rotation_matrix(m, &m1);
  multiply_rotation_matrices(m, &m2, &m1);

  return(1);
}

/*****************************************************************************/

int multiply_rotation_matrices(RotationMatrix *m,
			       RotationMatrix *a,
			       RotationMatrix *b)
{
  m->r11 = (a->r11*b->r11)+(a->r12*b->r21)+(a->r13*b->r31);
  m->r12 = (a->r11*b->r12)+(a->r12*b->r22)+(a->r13*b->r32);
  m->r13 = (a->r11*b->r13)+(a->r12*b->r23)+(a->r13*b->r33);

  m->r21 = (a->r21*b->r11)+(a->r22*b->r21)+(a->r23*b->r31);
  m->r22 = (a->r21*b->r12)+(a->r22*b->r22)+(a->r23*b->r32);
  m->r23 = (a->r21*b->r13)+(a->r22*b->r23)+(a->r23*b->r33);

  m->r31 = (a->r31*b->r11)+(a->r32*b->r21)+(a->r33*b->r31);
  m->r32 = (a->r31*b->r12)+(a->r32*b->r22)+(a->r33*b->r32);
  m->r33 = (a->r31*b->r13)+(a->r32*b->r23)+(a->r33*b->r33);

  return(1);
}

/*****************************************************************************/

int convert_r_to_q(RotationMatrix *m, Quaternion *q)
{
  double r11, r12, r13, r21, r22, r23, r31, r32, r33;
  double sxyz2[4], mag;
  int largest;

  r11 = m->r11;
  r12 = m->r12;
  r13 = m->r13;

  r21 = m->r21;
  r22 = m->r22;
  r23 = m->r23;

  r31 = m->r31;
  r32 = m->r32;
  r33 = m->r33;

  sxyz2[0] = 0.25*(1. + r11 + r22 + r33);
  sxyz2[1] = 0.25*(1. + r11 - r22 - r33);
  sxyz2[2] = 0.25*(1. - r11 + r22 - r33);
  sxyz2[3] = 0.25*(1. - r11 - r22 + r33);
  
  largest = ind_max_vec(&sxyz2[0], 4);

  switch(largest)
    {
    case 1:
      {
	q->x = sqrt(sxyz2[largest]);
	q->s = 0.25*(r32 - r23)/q->x;
	q->y = 0.25*(r12 + r21)/q->x;
	q->z = 0.25*(r13 + r31)/q->x;
	break;
      }
    case 2:
      {
	q->y = sqrt(sxyz2[largest]);
	q->s = 0.25*(r13 - r31)/q->y;
	q->x = 0.25*(r12 + r21)/q->y;
	q->z = 0.25*(r23 + r32)/q->y;
	break;
      }
    case 3:
      {
	q->z = sqrt(sxyz2[largest]);
	q->s = 0.25*(r21 - r12)/q->z;
	q->x = 0.25*(r13 + r31)/q->z;
	q->y = 0.25*(r23 + r32)/q->z;
	break;
      }
    default:
      {
	q->s = sqrt(sxyz2[largest]);
	q->x = 0.25*(r32 - r23)/q->s;
	q->y = 0.25*(r13 - r31)/q->s;
	q->z = 0.25*(r21 - r12)/q->s;
	break;
      }
    }

  mag = sqrt(q->s*q->s + q->x*q->x + q->y*q->y + q->z*q->z);
  q->s /= mag;
  q->x /= mag;
  q->y /= mag;
  q->z /= mag;

  return(1);
}
/*****************************************************************************/

int convert_q_to_r(Quaternion *q, RotationMatrix *m)
{
  double s, x, y, z;

  s = q->s;
  x = q->x;
  y = q->y;
  z = q->z;

  m->r11 = 1. - 2.*(y*y + z*z);
  m->r12 = 2.*(x*y - s*z);
  m->r13 = 2.*(x*z + s*y);

  m->r21 = 2.*(x*y + s*z);
  m->r22 = 1. - 2.*(x*x + z*z);
  m->r23 = 2.*(y*z - s*x);

  m->r31 = 2.*(x*z - s*y);
  m->r32 = 2.*(y*z + s*x);
  m->r33 = 1. - 2.*(x*x + y*y);

  return(1);
}

/*****************************************************************************/

int normalize_rotation_matrix(RotationMatrix *m)
{
  Quaternion q;

  convert_r_to_q(m, &q);
  convert_q_to_r(&q, m);

  return(1);
}

/*****************************************************************************/

/*****************************************************************************/

double length_of_line(DPOINT a, DPOINT b)
{
  double d;

  //d = sqrt((fabs(a.x - b.x) + 1.)*(fabs(a.x - b.x) + 1.) + 
  //(fabs(a.y - b.y) + 1.)*(fabs(a.y - b.y) + 1.));

  d = sqrt((a.x - b.x)*(a.x - b.x) + 
	   (a.y - b.y)*(a.y - b.y));

  return(d);
}

/*****************************************************************************/

double area_of_triangle_from_sides(double a, double b, double c)
{
  double area;
  
  area =  ( a + b + c);
  area *= (-a + b + c);
  area *= ( a - b + c);
  area *= ( a + b - c);
  area = 0.25*sqrt(area);

  return(area);
}

/*****************************************************************************/

double area_of_triangle(DPOINT v1, DPOINT v2, DPOINT v3)
{
  double a, b, c, area;
  
  a = length_of_line(v1, v2);
  b = length_of_line(v2, v3);
  c = length_of_line(v3, v1);

  area = area_of_triangle_from_sides(a, b, c);
  
  return(area);
}

/*****************************************************************************/

double area_of_four_sided_polygon(DPOINT v1, DPOINT v2, DPOINT v3, DPOINT v4)
{
  double area;

  area = area_of_triangle(v1, v2, v3);
  area += area_of_triangle(v3, v4, v1);

  return(area);
}

/*****************************************************************************/

#ifndef N_CONVEX_MAX 
#define N_CONVEX_MAX 20
#endif

int compute_convex_hull(DPOINT *out, DPOINT *p, int n)
{
  int i, flag, tst, n_v;
  double max_x = -1000.;
  DPOINT centroid;
  VERTIX v[N_CONVEX_MAX];
  VERTIX *v1, *w1, *start;

  centroid.x = 0.;
  centroid.y = 0.;

  /* data conversion to doubly linked list & computation of 
     centroid */
  
  for(i = 0; i < n; i++) {
    v[i].p.x = p[i].x;
    v[i].p.y = p[i].y;
    centroid.x += p[i].x;
    centroid.y += p[i].y;
  }

  centroid.x /= ((double) n);
  centroid.y /= ((double) n);

  /* compute angle for each point for sorting operation */

  for(i = 0; i < n; i++)
    v[i].angle = compute_convex_angle(&centroid, &v[i].p);

  /* sort by angle */

  sort_convex_angles(&v[0], n);

  /* compute starting vertex (largest x-value) */

  for(i = 0; i < n; i++) {
    if(v[i].p.x > max_x) {
      start = &v[i];
      max_x = v[i].p.x;
    }

    v[i].label = i+1; // for debugging purposes
  }

  /* create links to predecessors and successors */

  for(i = 1; i < n - 1; i++) {
    v[i].pred = &v[i-1];
    v[i].next = &v[i+1];
  }

  v[0].pred = &v[n-1];
  v[0].next = &v[1];

  v[n-1].pred = &v[n-2];
  v[n-1].next = &v[0];

  /* Graham Scan Convex Hull algorithm */

  flag = 0;
  v1 = start;
  w1 = v1->pred;
  n_v = n;

  while(v1->next != start || flag == 0) {
    if(v1->next == w1)
      flag = 1;
    tst = left_hand_turn(v1, v1->next, v1->next->next);
    
    //fprintf(stdout, "Checked %d %d %d --> tst = %d\n",
    //v1->label, v1->next->label, v1->next->next->label,tst);

    if(tst) {
      v1 = v1->next;
    } else {
      v1->next = v1->next->next;
      v1->next->pred = v1;
      v1 = v1->pred;
    }
  }

  /* convert data back to list of DPOINT's */
  
  v1 = start;
  n_v = 0;
  flag = 0;

  do {
    out[n_v].x = v1->p.x;
    out[n_v].y = v1->p.y;
    v1 = v1->next;
    n_v++;
  } while (v1 != start);

  return(n_v);
}

/*****************************************************************************/

double compute_convex_angle(DPOINT *centroid, DPOINT *p)
{
  double tmp;

  tmp = atan2(p->y - centroid->y, p->x - centroid->x);

  return(tmp);
}

/*****************************************************************************/

int sort_convex_angles(VERTIX *v, int n)
{
  int i, j;
  VERTIX x;

  for(j = 1; j < n; j++) {
    x.angle = v[j].angle;
    x.p.x = v[j].p.x;
    x.p.y = v[j].p.y;
    x.label = v[j].label;
    i = j - 1;
    while(i >= 0 && v[i].angle > x.angle) {
      v[i+1].angle = v[i].angle;
      v[i+1].p.x = v[i].p.x;
      v[i+1].p.y = v[i].p.y;
      v[i+1].label = v[i].label;
      i--;
    }
    v[i+1].angle = x.angle;
    v[i+1].p.x = x.p.x;
    v[i+1].p.y = x.p.y;
    v[i+1].label = x.label;
  }

  return(1);
}

/*****************************************************************************/

int left_hand_turn(VERTIX *v1, VERTIX *v2, VERTIX *v3)
{
  DPOINT p1, p2;

  p1.x = v2->p.x - v1->p.x;
  p1.y = v2->p.y - v1->p.y;

  p2.x = v3->p.x - v2->p.x;
  p2.y = v3->p.y - v2->p.y;

  if(p1.x * p2.y - p2.x * p1.y >= 0.)
    return(1);
  else
    return(0);
}

/*****************************************************************************/

double compute_area_of_convex_hull(DPOINT *p, int n)
{
  int i;
  double area;

  area = 0.;

  for(i = 1; i < n - 1; i++)
    area += area_of_triangle(p[0], p[i], p[i+1]);
  
  return(area);
}

/*****************************************************************************/

int convert_rotation_matrix(RotationMatrix *m, double r[3][3])
{
  m->r11 = r[0][0];
  m->r12 = r[0][1];
  m->r13 = r[0][2];

  m->r21 = r[1][0];
  m->r22 = r[1][1];
  m->r23 = r[1][2];

  m->r31 = r[2][0];
  m->r32 = r[2][1];
  m->r33 = r[2][2];

  return(1);
}

/*****************************************************************************/

int compute_zyx_euler_angles(double *abg, RotationMatrix *m)
{
  double a, b, g, cb;

  b = atan2(-(m->r31), sqrt((m->r11)*(m->r11) + (m->r21)*(m->r21)));

  if(is_zero2(b-M_PI/2., .001)) {
    b = M_PI/2.;
    a = 0.;
    g = atan2(m->r12, m->r22);
  } else if(is_zero2(b+M_PI/2., .001)) {
    b = -M_PI/2.;
    a = 0.;
    g = -atan2(m->r12, m->r22);
  } else {
    cb = cos(b);
    a = atan2(m->r21/cb, m->r11/cb);
    g = atan2(m->r32/cb, m->r33/cb);
  }

  abg[0] = a;
  abg[1] = b;
  abg[2] = g;

  return(1);
}

/*****************************************************************************/
