/*****************************************************************************/
/*                                                                           */
/* Author: Michael C. Nechyba                                                */
/* Last modified: 04/15/2001                                                 */
/*                                                                           */
/* File: "image.c"                                                           */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef INCLUDE_JAVA
#include "GatorCameraClient.h"
#endif

#include "acticast_include.h"

/****************************** Subroutines **********************************/

int convert_char(char c)
{
  int tmp;
  
  tmp = (int) c;
  tmp = tmp < 0 ? tmp + 256 : tmp;

  return(tmp);
}

/*****************************************************************************/

int read_ppm_file(char *file, PPM *im, int xdim, int ydim, int n_header)
{
  int i, j, k;
  FILE *fp;
  char tmp;
  
  im->xdim = xdim;
  im->ydim = ydim;
  
  fp = fopen(file, "r");
  
  if(fp == NULL) {
    fprintf(stderr, "Could not read %s file.\n", file);
    exit(-1);
  }
  
  for(i = 0; i < n_header; i++) {
    fscanf(fp, "%c", &tmp);
  }
  
  for(i = 0; i < im->ydim; i++) {
    for(j = 0; j < im->xdim; j++) {
      for(k = 0; k < 3; k++) {				
	fread(&im->p[j][i][k], sizeof(char), 1, fp);
      }
    }
  }
  
  fclose(fp);
  
  set_dimensions(im, 0, xdim, 0, ydim);

  return(1);	
}

/*****************************************************************************/

int write_ppm_file(char *file, PPM *im)
{
  int i, j, k;
  FILE *fp;
  
  fp = fopen(file, "w");
  if(fp == NULL) {
    fprintf(stderr, "Could not open %s for writing.\n", file);
    exit(1);
  }		
  
  fprintf(fp, "P6\n");
  fprintf(fp, "%d %d\n", im->xdim, im->ydim);
  fprintf(fp, "255\n");
  
  
  for(i = 0; i < im->ydim; i++) {
    for(j = 0; j < im->xdim; j++) {
      for(k = 0; k < 3; k++) {				
	fwrite(&im->p[j][i][k], sizeof(char), 1, fp);
      }
    }
  }
  
  fclose(fp);
  
  return(1);	
}

/*****************************************************************************/

int write_ppm_file_from_raw(char *file, char *im, int xdim, int ydim)
{
  int i, j, k;
  FILE *fp;
  char *p;
  
  fp = fopen(file, "w");
  if(fp == NULL) {
    fprintf(stderr, "Could not open %s for writing.\n", file);
    exit(1);
  }		
  
  fprintf(fp, "P6\n");
  fprintf(fp, "%d %d\n", xdim, ydim);
  fprintf(fp, "255\n");
  
  p = im;
  for(i = 0; i < ydim; i++) {
    for(j = 0; j < xdim; j++) {
      fprintf(fp, "%c", p[2]);
      fprintf(fp, "%c", p[1]);
      fprintf(fp, "%c", p[0]);
      p+=4;
    }
  }
  
  fclose(fp);
  
  return(1);	
}

/*****************************************************************************/

int write_ppm_file_small(char *file, PPM *im)
{
  int i, j, k, xdim, ydim;
  FILE *fp;
  
  fp = fopen(file, "w");
  if(fp == NULL) {
    fprintf(stderr, "Could not open %s for writing.\n", file);
    exit(1);
  }		
  
  xdim = im->xmax - im->xmin;
  ydim = im->ymax - im->ymin;
  
  fprintf(fp, "P6\n");
  fprintf(fp, "%d %d\n", xdim, ydim);
  fprintf(fp, "255\n");
  
  
  for(i = im->ymin; i < im->ymax; i++) {
    for(j = im->xmin; j < im->xmax; j++) {
      for(k = 0; k < 3; k++) {				
	fprintf(fp, "%c", im->p[j][i][k]);
      }
    }
  }
  
  fclose(fp);
  
  return(1);	
}

/*****************************************************************************/

int pixel_value(double *p, CAMERA_MODEL *g, double *v)
{
  double f, s11, s12, s13, s14, s21, s22, s23, s24, s31, s32, s33, s34;
  double x, y, z;
  
  f = g->f;
  
  s11 = g->s[0][0];
  s12 = g->s[0][1];
  s13 = g->s[0][2];
  s14 = g->s[0][3];
  s21 = g->s[1][0];
  s22 = g->s[1][1];
  s23 = g->s[1][2];
  s24 = g->s[1][3];
  s31 = g->s[2][0];
  s32 = g->s[2][1];
  s33 = g->s[2][2];
  s34 = g->s[2][3];
  
  x = v[0];
  y = v[1];
  z = v[2];
  
  p[0] = f*(s11*x + s12*y + s13*z + s14)/(s31*x + s32*y + s33*z + s34);
  p[1] = f*(s21*x + s22*y + s23*z + s24)/(s31*x + s32*y + s33*z + s34);
  
  return(1);
}

/*****************************************************************************/

int project_point(POINT *p, CAMERA_MODEL *camera, double *xyz)
{
  double tmp[2];
  
  pixel_value(&tmp[0], camera, xyz);
  
  p->x = (int) (tmp[0] + .5);
  p->y = (int) (tmp[1] + .5);
  
  return(1);
}

/*****************************************************************************/

int define_smoothing_filter(FILTER *f, int size)
{
  int i, j, n;
  
  f->size = size;
  
  n = 2*size + 1;

  for(i = 0; i < n; i++)
    for(j = 0; j < n; j++)
      f->m[i][j] = 1.;
  
  return(1);
}

/*****************************************************************************/

int filter(PPM *imin, PPM *imout, FILTER *f)
{
  int i, j, k, l, fsize, tmp;
  double r, g, b, w;	
  
  fsize = f->size;
  
  limit_dimensions(imin);
  copy_dimensions(imin, imout);
  
  for(i = imin->ymin; i < imin->ymax; i++) {
    for(j = imin->xmin; j < imin->xmax; j++) {
      r = 0.;
      g = 0.;
      b = 0.;
      w = 0.;
      for(k = -fsize; k <= fsize; k++) {
	for(l = -fsize; l <= fsize; l++) {
	  if((i+k) >= imin->ymin && (i+k) < imin->ymax && 
	     (j+l) >= imin->xmin && (j+l) < imin->xmax) {
	    tmp = convert_char(imin->p[j+l][i+k][0]);
	    r += (f->m[l+fsize][k+fsize]*((double) tmp));
	    
	    tmp = convert_char(imin->p[j+l][i+k][1]);
	    g += (f->m[l+fsize][k+fsize]*((double) tmp));
	    
	    tmp = convert_char(imin->p[j+l][i+k][2]);
	    b += (f->m[l+fsize][k+fsize]*((double) tmp));
	    
	    w += f->m[l+fsize][k+fsize];
	  }
	}
      }
      
      imout->p[j][i][0] = (int) (r/w + .5);
      imout->p[j][i][1] = (int) (g/w + .5);
      imout->p[j][i][2] = (int) (b/w + .5);
    }
  }
  
  return(1);
}

/*****************************************************************************/

int threshold(PPM *imin, PPM *imout, int t)
{
  int i, j, k, tmp, v;
  limit_dimensions(imin);
  copy_dimensions(imin, imout);
  
  for(i = imin->ymin; i < imin->ymax; i++) {
    for(j = imin->xmin; j < imin->xmax; j++) {
      tmp = 0;
      for(k = 0; k < 3; k++) {
	tmp += convert_char(imin->p[j][i][k]);
      }
      tmp /= 3;
      if(tmp < t) 
	v = 0;
      else 
	v = 255;
      
      imout->p[j][i][0] = (char) v;
      imout->p[j][i][1] = (char) v;
      imout->p[j][i][2] = (char) v;
    }
  }
  
  return(1);
}

/*****************************************************************************/

int set_dimensions(PPM *im, int xmin, int xmax, int ymin, int ymax)
{
  im->xmin = xmin;
  im->xmax = xmax;
  im->ymin = ymin;
  im->ymax = ymax;
  
  return(1);
}

/*****************************************************************************/

int copy_dimensions(PPM *imin, PPM *imout)
{
  imout->xdim = imin->xdim;
  imout->ydim = imin->ydim;
  imout->xmin = imin->xmin;
  imout->xmax = imin->xmax;
  imout->ymin = imin->ymin;
  imout->ymax = imin->ymax;

  return(1);
}

/*****************************************************************************/

int limit_dimensions(PPM *im)
{
  if(im->xmin < 0)
    im->xmin = 0;
  
  if(im->xmax > im->xdim-1)
    im->xmax = im->xdim-1;
  
  if(im->ymin < 0)
    im->ymin = 0;
  
  if(im->ymax > im->ydim-1)
    im->ymax = im->ydim-1;
  
  return(1);
}

/*****************************************************************************/

char difference_one_third_pixel(char c2, char c1)
{
  int x;
  char out;

  x = (convert_char(c2) - convert_char(c1));

#ifdef DIFFERENCE_HACK
  if(x < 0)
    x = 0;
#else
  x = abs(x);
#endif

  out = (char) (x);

  return(out);
}

/*****************************************************************************/

int difference(PPM *im2, PPM *im1, PPM *imout)
{
  int i, j, k;
  
  limit_dimensions(im1);
  copy_dimensions(im1, imout);
	
  for(i = im1->ymin; i < im1->ymax; i++) {
    for(j = im1->xmin; j < im1->xmax; j++) {
      for(k = 0; k < 3; k++) {
	imout->p[j][i][k] = 
	  difference_one_third_pixel(im2->p[j][i][k], im1->p[j][i][k]);
      }
    }
  }
  
  return(1);
}

/*****************************************************************************/

int copy_ppm_image(PPM *imin, PPM *imout)
{
  int i, j, k;
  
  limit_dimensions(imin);
  copy_dimensions(imin, imout);
  
  for(i = 0; i < imin->ydim; i++) {
    for(j = 0; j < imin->xdim; j++) {
      for(k = 0; k < 3; k++) {
	imout->p[j][i][k] = imin->p[j][i][k];
      }
    }
  }
  
  return(1);
}

/*****************************************************************************/

int copy_ppm_image_small(PPM *imin, PPM *imout)
{
  int i, j, k;
  
  limit_dimensions(imin);
  copy_dimensions(imin, imout);
  
  for(i = imin->ymin; i < imin->ymax; i++) {
    for(j = imin->xmin; j < imin->xmax; j++) {
      for(k = 0; k < 3; k++) {
	imout->p[j][i][k] = imin->p[j][i][k];
      }
    }
  }
  
  return(1);
}

/*****************************************************************************/

int make_single_color_image(PPM *im, int r, int g, int b)
{
  int i, j;
  
  for(i = 0; i < im->ydim; i++) {
    for(j = 0; j < im->xdim; j++) {
      im->p[j][i][0] = r;
      im->p[j][i][1] = g;
      im->p[j][i][2] = b;
    }
  }
  
  return(1);
}
	
/*****************************************************************************/

int draw_big_point_from_double(PPM *im, DPOINT *p, int r, int g, int b)
{
  POINT ip;

  ip.x = (int) (p->x + 0.5);
  ip.y = (int) (p->y + 0.5);

  if(ip.x >= 0 && ip.x < X_DIM && ip.y >= 0 && ip.y < Y_DIM)
    draw_big_point(im, &ip, r, g, b);

  return(1);
}

/*****************************************************************************/

int draw_big_point(PPM *im, POINT *p, int r, int g, int b)
{
  int x, y;
  
  x = p->x;
  y = p->y;
  
  if(x >= 0 && x < im->xdim && y >= 0 && y < im->ydim) {
    im->p[x][y][0] = (char) r;
    im->p[x][y][1] = (char) g;
    im->p[x][y][2] = (char) b;
  } else {
    return(-1);
  }
  
  if(x >= 1) {
    im->p[x-1][y][0] = (char) r;
    im->p[x-1][y][1] = (char) g;
    im->p[x-1][y][2] = (char) b;
  }
  
  if(x < (im->xdim-1)) {
    im->p[x+1][y][0] = (char) r;
    im->p[x+1][y][1] = (char) g;
    im->p[x+1][y][2] = (char) b;
  }
  
  if(y >= 1) {
    im->p[x][y-1][0] = (char) r;
    im->p[x][y-1][1] = (char) g;
    im->p[x][y-1][2] = (char) b;
  }
  
  if(y < (im->ydim-1)) {
    im->p[x][y+1][0] = (char) r;
    im->p[x][y+1][1] = (char) g;
    im->p[x][y+1][2] = (char) b;
  }
  
  return(1);
}

/*****************************************************************************/

int draw_point(PPM *im, POINT *p, int r, int g, int b)
{
  int x, y;
  
  x = p->x;
  y = p->y;
  
  if(x >= 0 && x < im->xdim && y >= 0 && y < im->ydim) {
    im->p[x][y][0] = (char) r;
    im->p[x][y][1] = (char) g;
    im->p[x][y][2] = (char) b;
  }
  return(1);
}

/*****************************************************************************/

int draw_line_from_double(PPM *im, DPOINT *p1, DPOINT *p2,
			  int r, int g, int b)
{
  double dx, dy;
  int xmin, ymin, xmax, ymax;

  dx = fabs(p1->x - p2->x);
  dy = fabs(p1->y - p2->y);

  if(dx > dy) {
    if(p1->x < p2->x) {
      xmin = (int) (p1->x + 0.5);
      ymin = (int) (p1->y + 0.5);

      xmax = (int) (p2->x + 0.5);
      ymax = (int) (p2->y + 0.5);
    } else {
      xmin = (int) (p2->x + 0.5);
      ymin = (int) (p2->y + 0.5);

      xmax = (int) (p1->x + 0.5);
      ymax = (int) (p1->y + 0.5);
    }
  } else {
    if(p1->y < p2->y) {
      xmin = (int) (p1->x + 0.5);
      ymin = (int) (p1->y + 0.5);

      xmax = (int) (p2->x + 0.5);
      ymax = (int) (p2->y + 0.5);
    } else {
      xmin = (int) (p2->x + 0.5);
      ymin = (int) (p2->y + 0.5);

      xmax = (int) (p1->x + 0.5);
      ymax = (int) (p1->y + 0.5);
    }
  }

  draw_line(im, xmin, xmax, ymin, ymax, r, g, b);

  return(1);
}

/*****************************************************************************/

int draw_line(PPM *im, int xmin, int xmax, int ymin, int ymax, 
	      int r, int g, int b)
{
  int x, y;
  double dx, dy, m;

  dx = ((double) (xmax - xmin));
  dy = ((double) (ymax - ymin));
  
  if(dx > dy) {

    m = dy/dx;
    
    for(x = xmin; x < xmax; x++) {
      y = (int) (((double) ymin) + m*((double) (x - xmin)) + .5);
      
      if(x >= 0 && x < X_DIM && y >= 0 && y < Y_DIM) {
	im->p[x][y][0] = (char) r;
	im->p[x][y][1] = (char) g;
	im->p[x][y][2] = (char) b;
      }
    }
  } else {
    m = dx/dy;
    
    for(y = ymin; y < ymax; y++) {
      x = (int) (((double) xmin) + m*((double) (y - ymin)) + .5);
      
      if(x >= 0 && x < X_DIM && y >= 0 && y < Y_DIM) {
	im->p[x][y][0] = (char) r;
	im->p[x][y][1] = (char) g;
	im->p[x][y][2] = (char) b;
      }
    }
  }
  
  return(1);
}

/*****************************************************************************/

int draw_region_box(PPM *im, int r, int g, int b)
{
  int xmin, xmax, ymin, ymax;
  
  xmin = im->xmin;
  xmax = im->xmax;
  ymin = im->ymin;
  ymax = im->ymax;
  
  draw_line(im, xmin, xmin, ymin, ymax, r, g, b);
  draw_line(im, xmax, xmax, ymin, ymax, r, g, b);
  draw_line(im, xmin, xmax, ymin, ymin, r, g, b);
  draw_line(im, xmin, xmax, ymax, ymax, r, g, b);
  
  return(1);
}

/*****************************************************************************/

int draw_ellipse(PPM *im, ELLIPSE *ell, int r, int g, int b)
{
  POINT p;
  double angle, dangle;
  double xtmp, ytmp, ct, st;

  dangle = 0.01;

  ct = cos(ell->theta);
  st = sin(ell->theta);

  for(angle = 0.; angle < 2.*M_PI; angle+=dangle) {
    xtmp = ell->a*cos(angle);
    ytmp = ell->b*sin(angle);

    p.x = (int) (ct*xtmp - st*ytmp + ell->mean[0] + 0.5);
    p.y = (int) (st*xtmp + ct*ytmp + ell->mean[1] + 0.5);

    p.x = limit_int(p.x, 0, im->xdim-1);
    p.y = limit_int(p.y, 0, im->ydim-1);

    draw_big_point(im, &p, r, g, b);
  }

  return(1);
}

/*****************************************************************************/

int mask_image(PPM *imin, PPM *imout, PPM *t)
{
  int i, j;

  limit_dimensions(imin);
  copy_dimensions(imin, imout);

  for(i = imin->ymin; i < imin->ymax; i++) {
    for(j = imin->xmin; j < imin->xmax; j++) {
      if(t->p[j][i][0] == (char) 255) {
	imout->p[j][i][0] = imin->p[j][i][0];
	imout->p[j][i][1] = imin->p[j][i][1];
	imout->p[j][i][2] = imin->p[j][i][2];
      } else {
	imout->p[j][i][0] = 0;
	imout->p[j][i][1] = 0;
	imout->p[j][i][2] = 0;
      }
    }
  }

  return(1);
}

/*****************************************************************************/

char hue_distance(char hc1, char hc2)
{
  int h1, h2, diff1, diff2;

  h1 = convert_char(hc1);
  h2 = convert_char(hc2);

  if(h2 >= h1) {
    diff1 = h2 - h1;
    diff2 = 255 - h2 + h1;
  } else {
    diff1 = h1 - h2;
    diff2 = 255 - h1 + h2;
  }
   
  return((char) min_int(diff1, diff2));
}

/*****************************************************************************/

int convert_rgb_to_hsv(PPM *imin, PPM *imout)
{
  int i, j;

  limit_dimensions(imin);
  copy_dimensions(imin, imout);

  for(i = imin->ymin; i < imin->ymax; i++)
    for(j = imin->xmin; j < imin->xmax; j++)
      convert_rgb_to_hsv_pixel(&imin->p[j][i][0], &imout->p[j][i][0]);

  return(1);
}

/*****************************************************************************/

int blank_red_channel_ppm2(PPM2 *newppm2)
{
  int i, j;

  // blank only red channel since it's used for blob detection
  // change this to use masks
  
  for(j = 0; j < newppm2->im.xdim; j++)
    for (i = 0; i < newppm2->im.ydim; i++)
      { 
	newppm2->im.p[j][i][0] = 0;
	// blank all 3 channels for now...
	// newppm2->im.p[j][i][1] = 0;
	// newppm2->im.p[j][i][2] = 0;
      }


  return(1);
}


/*****************************************************************************/

int convert_rgb_to_hsv_blob2(BLOB2 *blobin, int blobcount)
{
  int i, j;
  double buf1[3], buf2[3];

  for(j = 0; j < blobcount; j++)
    for (i = 0; i < blobin[j].n; i++)
      { 
	buf1[0] = blobin[j].rgb_t[0][i];
	buf1[1] = blobin[j].rgb_t[1][i];
	buf1[2] = blobin[j].rgb_t[2][i];
	convert_rgb_to_hsv_pixel_double(&buf1[0], &buf2[0]);
	blobin[j].rgb_t[0][i] = buf2[0];
	blobin[j].rgb_t[1][i] = buf2[1];
	blobin[j].rgb_t[2][i] = buf2[2];
      }

  return(1);
}

/*****************************************************************************/

int convert_rgb_to_hsv_pixel_double(double *rgb, double *hsv)
{
  double r, g, b, h, s, v, rc, gc, bc;
  double rgbmax, rgbmin;
  
  r = rgb[0] / 255.;
  g = rgb[1] / 255.;
  b = rgb[2] / 255.;

  rgbmax = rgb_max(r, g, b);
  rgbmin = rgb_min(r, g, b);

  v = rgbmax;

  if(!is_zero2(rgbmax, 0.000001)) {
    s = (rgbmax - rgbmin)/rgbmax;
  } else {
    s = 0.;
  }

  if(is_zero2(s, 0.000001)) {
    h = 0.;
  } else {
    rc = (rgbmax - r)/(rgbmax - rgbmin);
    gc = (rgbmax - g)/(rgbmax - rgbmin);
    bc = (rgbmax - b)/(rgbmax - rgbmin);

    if(is_zero2(r - rgbmax, 0.000001)) {
      h = bc - gc;
    } else if (is_zero2(g - rgbmax, 0.000001)) {
      h = 2. + rc - bc;
    } else {
      h = 4. + gc - rc;
    }
  }
  
  h /= 6.;

  if(h < 0.)
    h += 1.;

  h = limit_real(h*255., 0., 255.);
  s = limit_real(s*255., 0., 255.);
  v = limit_real(v*255., 0., 255.);

  hsv[0] = (h + 0.5);
  hsv[1] = (s + 0.5);
  hsv[2] = (v + 0.5);

  return(1);
}

/*****************************************************************************/

int convert_rgb_to_hsv_pixel(char *rgb, char *hsv)
{
  double r, g, b, h, s, v, rc, gc, bc;
  double rgbmax, rgbmin;
  
  r = (double) convert_char(rgb[0])/255.;
  g = (double) convert_char(rgb[1])/255.;
  b = (double) convert_char(rgb[2])/255.;

  rgbmax = rgb_max(r, g, b);
  rgbmin = rgb_min(r, g, b);

  v = rgbmax;

  if(!is_zero2(rgbmax, 0.000001)) {
    s = (rgbmax - rgbmin)/rgbmax;
  } else {
    s = 0.;
  }

  if(is_zero2(s, 0.000001)) {
    h = 0.;
  } else {
    rc = (rgbmax - r)/(rgbmax - rgbmin);
    gc = (rgbmax - g)/(rgbmax - rgbmin);
    bc = (rgbmax - b)/(rgbmax - rgbmin);

    if(is_zero2(r - rgbmax, 0.000001)) {
      h = bc - gc;
    } else if (is_zero2(g - rgbmax, 0.000001)) {
      h = 2. + rc - bc;
    } else {
      h = 4. + gc - rc;
    }
  }
  
  h /= 6.;

  if(h < 0.)
    h += 1.;

  h = limit_real(h*255., 0., 255.);
  s = limit_real(s*255., 0., 255.);
  v = limit_real(v*255., 0., 255.);

  hsv[0] = (char) (h + 0.5);
  hsv[1] = (char) (s + 0.5);
  hsv[2] = (char) (v + 0.5);
  
  return(1);
}

/*****************************************************************************/

double rgb_max(double r, double g, double b)
{
  if(r >= g && r >= b)
    return(r);
  else if (g >= r && g >= b)
    return(g);
  else
    return(b);
}

/*****************************************************************************/

double rgb_min(double r, double g, double b)
{
  if(r <= g && r <= b)
    return(r);
  else if (g <= r && g <= b)
    return(g);
  else
    return(b);
}

/*****************************************************************************/

int difference_partial(PPM *im2, PPM *im1, PPM *imout, int which,
		       char (*diff_func)(char, char))
{
  int i, j, k;
  int s1, s2;

  limit_dimensions(im1);
  copy_dimensions(im1, imout);
	
  for(i = im1->ymin; i < im1->ymax; i++) {
    for(j = im1->xmin; j < im1->xmax; j++) {
      imout->p[j][i][which] = 
	diff_func(im2->p[j][i][which], im1->p[j][i][which]);

      /* saturation HACK */

      if(which == 1) {
	s1 = convert_char(im1->p[j][i][which]);
	s2 = convert_char(im2->p[j][i][which]);

	if(min_int(s1, s2) < 10) {
	  imout->p[j][i][0] = 0;
	}
      }
    }
  }
  
  return(1);
}

/*****************************************************************************/

int or_images(PPM *im1, PPM *im2, PPM *imout)
{
  int i, j;
  
  limit_dimensions(im1);
  copy_dimensions(im1, imout);
  
  for(i = im1->ymin; i < im1->ymax; i++) {
    for(j = im1->xmin; j < im1->xmax; j++) {
      if(im1->p[j][i][0] == (char) 255 ||
	 im2->p[j][i][0] == (char) 255) {
	imout->p[j][i][0] = (char) 255;
	imout->p[j][i][1] = (char) 255;
	imout->p[j][i][2] = (char) 255;
      } else {
	imout->p[j][i][0] = 0;
	imout->p[j][i][1] = 0;
	imout->p[j][i][2] = 0;
      }
    }
  }
  
  return(1);
}

/*****************************************************************************/

int and_images(PPM *im1, PPM *im2, PPM *imout)
{
  int i, j;
  
  limit_dimensions(im1);
  copy_dimensions(im1, imout);
  
  for(i = im1->ymin; i < im1->ymax; i++) {
    for(j = im1->xmin; j < im1->xmax; j++) {
      if(im1->p[j][i][0] == (char) 255 &&
	 im2->p[j][i][0] == (char) 255) {
	imout->p[j][i][0] = (char) 255;
	imout->p[j][i][1] = (char) 255;
	imout->p[j][i][2] = (char) 255;
      } else {
	imout->p[j][i][0] = 0;
	imout->p[j][i][1] = 0;
	imout->p[j][i][2] = 0;
      }
    }
  }
  
  return(1);
}

/*****************************************************************************/

int threshold_hsv(PPM *imin, PPM *imout, int th, int ts, int tv)
{
  int i, j, k, tmp, h, s, v, o;
  limit_dimensions(imin);
  copy_dimensions(imin, imout);
  
  for(i = imin->ymin; i < imin->ymax; i++) {
    for(j = imin->xmin; j < imin->xmax; j++) {
      h = convert_char(imin->p[j][i][0]);
      s = convert_char(imin->p[j][i][1]);
      v = convert_char(imin->p[j][i][2]);
      if(h < th && s < ts && v < tv) {
	o = 0;
      }	else {
	o = 255;
      }
      
      imout->p[j][i][0] = (char) o;
      imout->p[j][i][1] = (char) o;
      imout->p[j][i][2] = (char) o;
    }
  }
  
  return(1);
}

/*****************************************************************************/

int border_pixel(PPM *im, int x, int y, int label)
{
  int xl, xr, yu, yd;

  xl = max_int(im->xmin, x-1);
  xr = min_int(im->xmax, x+1);
  yu = max_int(im->ymin, y-1);
  yd = min_int(im->ymax, y+1);

  if((!((im->p[xl][y][0] == (char) label) &&
       (im->p[xr][y][0] == (char) label) &&
       (im->p[x][yu][0] == (char) label) &&
       (im->p[x][yd][0] == (char) label))) && im->p[x][y][0] == (char) label) {
    return(1);
  } else {
    return(-1);
  }
}

/*****************************************************************************/

int border_extraction(PPM *im, int x, int y, int label)
{
  if(border_pixel(im, x, y, label) < 0 || im->p[x][y][1] == (char) 128) {
    return(1);
  } else {
#ifndef SUPPRESS_ALL_OUTPUT
#ifdef LEVEL_THREE_DIAGNOSTIC
    fprintf(stderr, "Border:  %d  %d\n", x, y);
#endif
#endif
    im->p[x][y][1] = (char) 128;
  }

  if((y-1) >= im->ymin)
    border_extraction(im, x, y-1, label);

  if((x-1) >= im->xmin) 
    border_extraction(im, x-1, y, label);

  if((x+1) < im->xmax) 
    border_extraction(im, x+1, y, label);

  if((y+1) < im->ymax) 
    border_extraction(im, x, y+1, label);

  if((x-1) >= im->xmin && (y-1) >= im->ymin) 
    border_extraction(im, x-1, y-1, label);

  if((x+1) < im->xmax && (y-1) >= im->ymin) 
    border_extraction(im, x+1, y-1, label);

  if((x-1) >= im->xmin && (y+1) < im->ymax) 
    border_extraction(im, x-1, y+1, label);

  if((x+1) < im->xmax && (y+1) < im->ymax)
    border_extraction(im, x+1, y+1, label);

  return(1);
}

/*****************************************************************************/

/* NEW BLOB2 structure */

int allocate_blob2(BLOB2 *b, int n_max)
{
  initialize_blob2(b);

  /* Note that we are allocating rgb' and xy' (transposes) */
  /* This is more efficient than allocating rgb and xy     */

  b->rgb_t = allocate_matrix(b->rgb_t, 3, n_max);
  b->xy_t  = allocate_matrix(b->xy_t, 2, n_max);
  b->type  = int_allocate_vector(b->type, n_max);

  return(1);
}

/*****************************************************************************/

int initialize_blob2(BLOB2 *b)
{
  b->label = 0;
  b->n = 0;
  b->first.x = 0;
  b->first.y = 0;
  initialize_ellipse(&b->ell);

  return(1);
}

/*****************************************************************************/

int convert_double_char_to_int(char *two_bytes)
{
  int i, n, b[2];
  
  for(i = 0; i < 2; i++) {
    b[i] = (int) two_bytes[i];
    b[i] = b[i] < 0 ? b[i] + 256 : b[i];
  }

  n = b[0]*256 + b[1];
  return(n);
}

/*****************************************************************************/

int convert_int_to_double_char(char *two_bytes, int n)
{
  int n1, n2;

  if(n > 65535) {
    fprintf(stderr, "FATAL ERROR: label number is too large (%d).\n", n);
    exit(-1);
  }

  n1 = n/256;
  n2 = n%256;

  two_bytes[0] = (char) n1;
  two_bytes[1] = (char) n2;

  return(1);
}

/*****************************************************************************/

int detect_blobs_ppm2(RAW2 *frame, PPM2 *t, BLOB2 *b, int min_blob_size)
{
  int i, j, j1, count, label;
  
  count = 0;
  label = 1;
  
  limit_dimensions(&t->im);

  for(i = 0; i < t->im.ydim; i++) {
    for(j1 = 0; j1 < t->m.line[i].n; j1+=2) {      
      for(j = t->m.line[i].c[j1].x; j <= t->m.line[i].c[j1+1].x; j++) {
	if(t->im.p[j][i][0] == (char) 255) { 
	  initialize_blob2(&b[count]); 
	  b[count].first.x = j; 
	  b[count].first.y = i;
	  b[count].label = label;
	  collect_blob_data_ppm2(frame, t, j, i, label, 255, &b[count]);
	  if(b[count].n >= min_blob_size) {
	    //fprintf(stderr, "Blob %d: %d\n", label, b[count].n);
	    count++;
	    label++;
	  } else {
	    //fprintf(stderr, "Small Blob %d: %d\n", label, b[count].n);
	    collect_blob_data_ppm2(frame, t, 
				   b[count].first.x, b[count].first.y,
				   255, label, &b[count]);
	  }
	  
	  if(label >= 255) {
	    fprintf(stderr, "Found more than 254 blobs of size > %d.\n",
		    min_blob_size);
	    fprintf(stderr, "This is a FATAL error.\n");
	    exit(-1);
	  }
	}
      }
    }
  }

  for(i = 0; i < count; i++) {
    //    fprintf(stderr, ">>>>>>>>>>>>>>>>>>\n");
    compute_ellipse_from_online(&b[i].ell, b[i].n);
    //    display_ellipse(stderr, &b[i].ell);
    //    fprintf(stderr, "<<<<<<<<<<<<<<<<<<\n");
  }

#ifdef COLOR_DIAGNOSTIC
  for(i = 0; i < count; i++)
    for(j = 0; j < b[i].n; j++) 
    fprintf(stdout, "RGB: %d %.0f %.0f %.0f %.0f %.0f\n", 
            b[i].n,
            b[i].rgb_t[0][j],
            b[i].rgb_t[1][j],
            b[i].rgb_t[2][j],
	    b[i].xy_t[0][j],
	    b[i].xy_t[1][j]);
#endif

  return(count);
}

/*****************************************************************************/

int collect_blob_data_ppm2(RAW2 *frame, PPM2 *t, int x, int y, 
			   int label, int tst, BLOB2 *b)
{
  int loc;
  double xd, yd;
  
  xd = (double) x;
  yd = (double) y;

  loc = (y*frame->im.xdim + x)*4;

  t->im.p[x][y][0] = (char) label;
  b->rgb_t[RED][b->n] =   (double) convert_char(frame->im.p[loc+2]);
  b->rgb_t[GREEN][b->n] = (double) convert_char(frame->im.p[loc+1]);
  b->rgb_t[BLUE][b->n] =  (double) convert_char(frame->im.p[loc]);
  b->xy_t[0][b->n] = xd;
  b->xy_t[1][b->n] = yd;
  b->type[b->n] = 0;
  b->n++;

  //fprintf(stderr, "b->n = %d (%d)\n", b->n, label);
  //fprintf(stderr, "-------------------\n");
  //display_ellipse(stderr, &b->ell);

  b->ell.mean[0] += xd;
  b->ell.mean[1] += yd;
  b->ell.cov[0][0] += (xd*xd);
  b->ell.cov[0][1] += (yd*xd);
  b->ell.cov[1][0] = b->ell.cov[0][1];
  b->ell.cov[1][1] += (yd*yd);

  b->ell.xmin = min_int(b->ell.xmin, x);
  b->ell.xmax = max_int(b->ell.xmax, x);
  b->ell.ymin = min_int(b->ell.ymin, y);
  b->ell.ymax = max_int(b->ell.ymax, y);

#ifdef EIGHT_CONNECTED  
  if((x-1) >= 0 && (y-1) >= 0) 
    if(t->im.p[x-1][y-1][0] == (char) tst) 
      collect_blob_data_ppm2(frame, t, x-1, y-1, label, tst, b);
#endif

  if((y-1) >= 0)
    if(t->im.p[x][y-1][0] == (char) tst) 
      collect_blob_data_ppm2(frame, t, x, y-1, label, tst, b);

#ifdef EIGHT_CONNECTED
  if((x+1) < t->im.xdim && (y-1) >= 0) 
    if(t->im.p[x+1][y-1][0] == (char) tst) 
      collect_blob_data_ppm2(frame, t, x+1, y-1, label, tst, b);
#endif

  if((x-1) >= 0) 
    if(t->im.p[x-1][y][0] == (char) tst) 
      collect_blob_data_ppm2(frame, t, x-1, y, label, tst, b);

  if((x+1) < t->im.xdim) 
    if(t->im.p[x+1][y][0] == (char) tst) 
      collect_blob_data_ppm2(frame, t, x+1, y, label, tst, b);

#ifdef EIGHT_CONNECTED
  if((x-1) >= 0 && (y+1) < t->im.ydim) 
    if(t->im.p[x-1][y+1][0] == (char) tst) 
      collect_blob_data_ppm2(frame, t, x-1, y+1, label, tst, b);
#endif

  if((y+1) < t->im.ydim) 
    if(t->im.p[x][y+1][0] == (char) tst) 
      collect_blob_data_ppm2(frame, t, x, y+1, label, tst, b);

#ifdef EIGHT_CONNECTED
  if((x+1) < t->im.xdim && (y+1) < t->im.ydim)
    if(t->im.p[x+1][y+1][0] == (char) tst) 
     collect_blob_data_ppm2(frame, t, x+1, y+1, label, tst, b);
#endif

  return(1);
}

/*****************************************************************************/

int make_single_color_image_ppm2(PPM2 *t, int r, int g, int b)
{
  int i, j, j1, count, label;
  
  for(i = 0; i < t->im.ydim; i++) {
    for(j1 = 0; j1 < t->m.line[i].n; j1+=2) {      
      for(j = t->m.line[i].c[j1].x; j <= t->m.line[i].c[j1+1].x; j++) {
	t->im.p[j][i][0] = r;
	t->im.p[j][i][1] = g;
	t->im.p[j][i][2] = b;
      }
    }
  }
  
  return(1);
}
	
/*****************************************************************************/

int extract_blobs2_init(BLOB2 *blobs, RAW2 *im, RAW2 *ref, 
			PPM2 *tmp1, PPM2 *tmp2, PPM2 *tmp3,
			BOUNDING_BOX *b, int n, int threshold)
{
  int count = 0;

  difference_raw2(im, ref, tmp1);
  //write_ppm_file("tmp1.ppm", &tmp1->im);
  median_filter_ppm2(tmp1, tmp2, 1, threshold - 5);
  //write_ppm_file("tmp2.ppm", &tmp2->im);
  threshold_ppm2(tmp2, tmp3, threshold);
  //write_ppm_file("tmp3.ppm", &tmp3->im);
  count = detect_blobs_ppm2(im, tmp3, blobs, MIN_BLOB_SIZE);

  return(count);
}

/*****************************************************************************/

int extract_blobs2(BLOB2 *blobs, RAW2 *im, RAW2 *ref, 
		   PPM2 *tmp1, PPM2 *tmp2, PPM2 *tmp3,
		   BOUNDING_BOX *b, int n, int threshold)
{
  int count = 0;

  dilate_bounding_boxes(b, n);                      // no ivan
  create_mask(&ref->m, b, n);
  difference_raw2(im, ref, tmp1);
  median_filter_ppm2(tmp1, tmp2, 1, threshold - 5); // no ivan
  copy_mask_ppm2_to_ppm2(tmp2, tmp3);               // no ivan
  make_single_color_image_ppm2(tmp3, 0, 0, 0);      // no ivan 
  contract_bounding_boxes(b, n);                    // no ivan
  create_mask(&tmp2->m, b, n);                      // no ivan
  threshold_ppm2(tmp2, tmp3, threshold);
  count = detect_blobs_ppm2(im, tmp3, blobs, MIN_BLOB_SIZE);

  return(count);
}

/*****************************************************************************/

int display_blob2(FILE *fp, BLOB2 *b)
{
  int i;

  for(i = 0; i < b->n; i++)
    fprintf(fp, "%d %d %d %d %d %d\n",
	    (int) b->rgb_t[RED][i],
	    (int) b->rgb_t[GREEN][i],
	    (int) b->rgb_t[BLUE][i],
	    (int) b->xy_t[0][i],
	    (int) b->xy_t[1][i],
	    b->type[i]);

  return(1);
}

/*****************************************************************************/

int merge_blobs(BLOB2 *out, BLOB2 *blist, int n)
{
  int i, j, count = 0;
  
  for(i = 0; i <n; i++)
    for(j = 0; j < blist[i].n; j++) {
      out->rgb_t[RED][count]   = blist[i].rgb_t[RED][j];
      out->rgb_t[GREEN][count] = blist[i].rgb_t[GREEN][j];
      out->rgb_t[BLUE][count]  = blist[i].rgb_t[BLUE][j];
      out->xy_t[0][count]   = blist[i].xy_t[0][j];
      out->xy_t[1][count] = blist[i].xy_t[1][j];
      count++;
    }

  out->n = count;

  compute_ellipse(&out->ell, out->xy_t, out->n,
		  CONFIDENCE_INTERVAL, 1);
    
  return(count);
}

/*****************************************************************************/

int draw_from_file(char *draw_file, PPM *im, RAW *im_raw)
{
  FILE *fp;
  int type, r, g, b;
  BOUNDING_BOX box;
  ELLIPSE ell;
  POINT p, min, max;

  if(file_exist(draw_file) == 0) {
    fprintf(stderr, "Could not read draw file %s. Not drawing.\n", draw_file);
    return(-1);
  }

  fp = fopen(draw_file, "r");

  while(fscanf(fp, "%d", &type) != EOF) {
    fscanf(fp, "%d", &r);
    fscanf(fp, "%d", &g);
    fscanf(fp, "%d", &b);

    switch(type)
      {
      case POINT_TYPE:
        {       
	  fscanf(fp, "%d", &p.x);
	  fscanf(fp, "%d", &p.y);
	  if(im_raw != NULL)
	    draw_point_raw(im_raw->p, &p, r, g, b);
	  if(im != NULL)
	    draw_point(im, &p, r, g, b);
          break;
        }
      case BIG_POINT_TYPE:
        {       
	  fscanf(fp, "%d", &p.x);
	  fscanf(fp, "%d", &p.y);
	  if(im_raw != NULL)
	    draw_big_point_raw(im_raw->p, &p, r, g, b);
          draw_big_point(im, &p, r, g, b);
          break;
        }
      case LINE_TYPE:
        {       
	  fscanf(fp, "%d", &min.x);
	  fscanf(fp, "%d", &min.y);
	  fscanf(fp, "%d", &max.x);
	  fscanf(fp, "%d", &max.y);
	  if(im_raw != NULL)
	    draw_line_raw(im_raw->p, min.x, max.x, min.y, max.y, r, g, b);
	  if(im != NULL)
	    draw_line(im, min.x, max.x, min.y, max.y, r, g, b);
          break;
        }
      case REGION_BOX_TYPE:
        {       
	  fscanf(fp, "%d", &im->xmin);
	  fscanf(fp, "%d", &im->ymin);
	  fscanf(fp, "%d", &im->xmax);
	  fscanf(fp, "%d", &im->ymax);
	  box.xmin = limit_int(im->xmin, 0, X_DIM-1);
	  box.xmax = limit_int(im->xmax, 0, X_DIM-1);
	  box.ymin = limit_int(im->ymin, 0, Y_DIM-1);
	  box.ymax = limit_int(im->ymax, 0, Y_DIM-1);
	  if(im_raw != NULL)
	    draw_region_box_raw(im_raw->p, r, g, b, &box);
	  if(im != NULL)
	    draw_region_box(im, r, g, b);
          break;
        }
      case ELLIPSE_TYPE:
        {       
	  fscanf(fp, "%lf", &ell.mean[0]);
	  fscanf(fp, "%lf", &ell.mean[1]);
	  fscanf(fp, "%lf", &ell.a);
	  fscanf(fp, "%lf", &ell.b);
	  fscanf(fp, "%lf", &ell.theta);
	  if(im_raw != NULL)
	    draw_ellipse_raw(im_raw->p, &ell, r, g, b);
	  if(im != NULL)
	    draw_ellipse(im, &ell, r, g, b);
          break;
        }
      }
  }  
  fclose(fp);

  return(1);
}

/*****************************************************************************/

int display_camera_model(FILE *fp, CAMERA_MODEL *c)
{
  fprintf(fp, "F %.15f\n", c->f);
  fprintf(fp, "S11 %.15f\n", c->s[0][0]);
  fprintf(fp, "S12 %.15f\n", c->s[0][1]);
  fprintf(fp, "S13 %.15f\n", c->s[0][2]);
  fprintf(fp, "S14 %.15f\n", c->s[0][3]);
  fprintf(fp, "S21 %.15f\n", c->s[1][0]);
  fprintf(fp, "S22 %.15f\n", c->s[1][1]);
  fprintf(fp, "S23 %.15f\n", c->s[1][2]);
  fprintf(fp, "S24 %.15f\n", c->s[1][3]);
  fprintf(fp, "S31 %.15f\n", c->s[2][0]);
  fprintf(fp, "S32 %.15f\n", c->s[2][1]);
  fprintf(fp, "S33 %.15f\n", c->s[2][2]);
  fprintf(fp, "S34 %.15f\n", c->s[2][3]);

  return(1);
}

/*****************************************************************************/

int write_camera_model(char *file_name, CAMERA_MODEL *c)
{
  FILE *fp;

  fp = fopen(file_name, "w");
  display_camera_model(fp, c);
  fclose(fp);
}

/*****************************************************************************/

// assumes that camera calibration file is in strict canonical order

int read_camera_model(char *file_name, CAMERA_MODEL *c)
{
  FILE *fp;
  char str[256];
  int i, j;

  file_no_exist(file_name, "camera model");
  
  fp = fopen(file_name, "r");

  fscanf(fp, "%s", &str[0]);
  fscanf(fp, "%lf", &c->f);

  for(i = 0; i < 3; i++)
    for(j = 0; j < 4; j++) {
        fscanf(fp, "%s", &str[0]);
	fscanf(fp, "%lf", &c->s[i][j]);
    }

  fclose(fp);

  return(1);
}

/*****************************************************************************/

/*
  This function converts a camera calibration model to a coordinate
  system scaled by scaleFactor
*/

int scale_camera_model(CAMERA_MODEL *c, double scaleFactor)
{
  c->s[0][0] *= scaleFactor;
  c->s[0][1] *= scaleFactor;
  c->s[0][2] *= scaleFactor;

  c->s[1][0] *= scaleFactor;
  c->s[1][1] *= scaleFactor;
  c->s[1][2] *= scaleFactor;

  c->s[2][0] *= scaleFactor;
  c->s[2][1] *= scaleFactor;
  c->s[2][2] *= scaleFactor;

  return(1);
}

/*****************************************************************************/

int compute_bounding_boxes_from_blobs(BLOB2 *blobs, int n_blobs,
				      BOUNDING_BOX *boxes, 
				      int expand_constant)
{
  int i;

  for(i = 0; i < n_blobs; i++) {
    boxes[i].xmin = max_int(0, blobs[i].ell.xmin - expand_constant); 
    boxes[i].xmax = min_int(X_DIM-1, blobs[i].ell.xmax + expand_constant); 
    
    boxes[i].ymin = max_int(0, blobs[i].ell.ymin - expand_constant); 
    boxes[i].ymax = min_int(Y_DIM-1, blobs[i].ell.ymax + expand_constant); 
  }

  return(n_blobs);
}

/*****************************************************************************/
/*****************************************************************************/

// QR decomposition adapted from Numerical Recipes in C

int qrdcmp(double **a, int n, double *c, double *d)
{
  int i, j, k, sing = 0;
  double scale = 0., sigma, sum, tau;
  
  for (k = 0; k < n-1; k++) {
    for (i = k; i < n; i++) 
      scale = max_real(scale, fabs(a[i][k]));

    if (scale == 0.0) {
      sing=1;
      c[k] = d[k] = 0.0;
    } else {
      for (i = k;i < n; i++) 
	a[i][k] /= scale;

      for (sum = 0.0, i = k; i < n; i++) 
	sum += a[i][k]*a[i][k];

      sigma = a[k][k] >= 0. ? fabs(sqrt(sum)) : -fabs(sqrt(sum));
      a[k][k] += sigma;
      c[k] = sigma*a[k][k];
      d[k] = -scale*sigma;

      for (j = k + 1;j < n; j++) {
	for (sum = 0.0, i = k; i < n; i++) 
	  sum += a[i][k]*a[i][j];
	tau = sum/c[k];
	for (i = k;i < n; i++) 
	  a[i][j] -= tau*a[i][k];
      }
    }
  }

  d[n-1] = a[n-1][n-1];

  if(d[n-1] == 0.0) 
    sing=1;

  return(sing);
}

/*****************************************************************************/

int convert_householder_to_q(double **q, int n, double *u, double c)
{
  int i, j;
  double c2;

  c2 = (-1./c);
  
  vector_to_matrix(q, u, n);
  
  for(i = 0; i < n; i++)
    for(j = 0; j < n; j++)
      q[i][j] *= c2;

  for(i = 0; i < n; i++)
    q[i][i] += 1.;

  return(1);
}

/*****************************************************************************/

int qr_decomposition(double **a, int n, double **q, double **r)
{
  int i, j, k;
  double *c, *d, **qt1, **qt2;

  /* memory allocation */

  c = allocate_vector(c, n - 1);
  d = allocate_vector(d, n);
  qt1 = allocate_matrix(qt1, n, n); 
  qt2 = allocate_matrix(qt1, n, n); 

  qrdcmp(a, n, c, d);
  
  zero_matrix(r, n, n);
  for(i = 0; i < n; i++)
    r[i][i] = d[i];

  for(i = 0; i < n; i++)
    for(j = i + 1; j < n; j++)
      r[i][j] = a[i][j];

  identity_matrix(q, n);

  for(i = 0; i < n - 1; i++) {
    for(j =  0; j < i; j++)
      d[j] = 0.;
    for(j = i; j < n; j++)
      d[j] = a[j][i];
    
    convert_householder_to_q(qt1, n, d, c[i]);
    multiply_two_matrices(qt2, q, n, n, qt1, n);
    copy_matrix(qt2, q, n, n);
  }

  /* memory de-allocation */

  free_matrix(qt2, n);
  free_matrix(qt1, n);
  free_vector(d);
  free_vector(c);

  return(1);
}

/*****************************************************************************/

int intrinsic_extrinsic_decomposition(CAMERA_MODEL *c, INTRINSIC_EXTRINSIC *d)
{
  int i, j;
  double **a, **b, **q, **r, trans[3];

  a = allocate_matrix(a, 3, 3);
  b = allocate_matrix(b, 3, 3);
  q = allocate_matrix(q, 3, 3);
  r = allocate_matrix(r, 3, 3);

  for(i = 0; i < 3; i++)
    for(j = 0; j < 3; j++)
      a[i][j] = c->s[i][j];

  trans[0] = c->s[0][3];
  trans[1] = c->s[1][3];
  trans[2] = c->s[2][3];

  inverse(b, a, 3);
  qr_decomposition(b, 3, q, r);
  inverse(a, r, 3);

  for(i = 0; i < 3; i++)
    for(j = 0; j < 3; j++) {
      a[i][j] /= a[2][2];
      d->k[i][j] = a[i][j];
    }

  for(i = 0; i < 3; i++)
    for(j = 0; j < 3; j++)
      d->r[i][j] = q[j][i];


  matrix_vector_product(&d->t[0], r, &trans[0], 3, 3);

  /*
  if(d->k[0][0] < 0.) {
    d->k[0][0] = -(d->k[0][0]);
    d->t[0] = -(d->t[0]);
    for(i = 0; i < 3; i++)
      d->r[0][i] = -(d->r[0][i]);
  }

  if(d->k[1][1] < 0.) {
    d->k[1][1] = -(d->k[1][1]);
    d->t[1] = -(d->t[1]);
    for(i = 0; i < 3; i++)
      d->r[1][i] = -(d->r[1][i]);
  }
  */
  //d->t[0] = -(d->t[0]);
  //d->t[1] = -(d->t[1]);
  //d->t[2] = -(d->t[2]);

  free_matrix(r, 3);
  free_matrix(q, 3);
  free_matrix(b, 3);
  free_matrix(a, 3);

  return(1);
}

/*****************************************************************************/

int display_intrinsic_extrinsic(FILE *fp, INTRINSIC_EXTRINSIC *d)
{
  int i, j;
  
  fprintf(fp, "K-matrix:\n\n");
  
  for(i = 0; i < 3; i++) {
    for(j = 0; j < 3; j++) {
      fprintf(fp, " %.10f", d->k[i][j]);
    }
    fprintf(fp, "\n");
  }

  fprintf(fp, "\nR-matrix:\n\n");

  for(i = 0; i < 3; i++) {
    for(j = 0; j < 3; j++) {
      fprintf(fp, " %.10f", d->r[i][j]);
    }
    fprintf(fp, "\n");
  }

  fprintf(fp, "\nT-vector:\n\n");
  
  display_vector(fp, &d->t[0], 3, 1, 10);

  return(1);
}

/*****************************************************************************/

int draw_convex_hull(PPM *im, DPOINT *pts, int n_pts, DPOINT *ch, int n_ch)
{
  int i;

  for(i = 0; i < n_pts; i++)
    draw_big_point_from_double(im, &pts[i], COLOR_GREEN);

  for(i = 0; i < n_ch - 1; i++)
    draw_line_from_double(im, &ch[i], &ch[i+1], COLOR_YELLOW);
  draw_line_from_double(im, &ch[n_ch-1], &ch[0], COLOR_YELLOW);

  for(i = 0; i < n_ch; i++)
    draw_big_point_from_double(im, &ch[i], COLOR_RED);
  
  return(1);
}

/*****************************************************************************/

int draw_line_from_double_raw(char *im, DPOINT *p1, DPOINT *p2,
			  int r, int g, int b)
{
  double dx, dy;
  int xmin, ymin, xmax, ymax;

  dx = fabs(p1->x - p2->x);
  dy = fabs(p1->y - p2->y);

  if(dx > dy) {
    if(p1->x < p2->x) {
      xmin = (int) (p1->x + 0.5);
      ymin = (int) (p1->y + 0.5);

      xmax = (int) (p2->x + 0.5);
      ymax = (int) (p2->y + 0.5);
    } else {
      xmin = (int) (p2->x + 0.5);
      ymin = (int) (p2->y + 0.5);

      xmax = (int) (p1->x + 0.5);
      ymax = (int) (p1->y + 0.5);
    }
  } else {
    if(p1->y < p2->y) {
      xmin = (int) (p1->x + 0.5);
      ymin = (int) (p1->y + 0.5);

      xmax = (int) (p2->x + 0.5);
      ymax = (int) (p2->y + 0.5);
    } else {
      xmin = (int) (p2->x + 0.5);
      ymin = (int) (p2->y + 0.5);

      xmax = (int) (p1->x + 0.5);
      ymax = (int) (p1->y + 0.5);
    }
  }

  draw_line_raw(im, xmin, xmax, ymin, ymax, r, g, b);

  return(1);
}

/*****************************************************************************/

int draw_big_point_from_double_raw(char *im, DPOINT *p, int r, int g, int b)
{
  POINT ip;

  ip.x = (int) (p->x + 0.5);
  ip.y = (int) (p->y + 0.5);

  if(ip.x >= 0 && ip.x < X_DIM && ip.y >= 0 && ip.y < Y_DIM)
    draw_big_point_raw(im, &ip, r, g, b);

  return(1);
}

/*****************************************************************************/

int draw_convex_hull_raw(char *im, DPOINT *pts, int n_pts, 
			 DPOINT *ch, int n_ch)
{
  int i;

  for(i = 0; i < n_pts; i++)
    draw_big_point_from_double_raw(im, &pts[i], COLOR_CYAN);

  for(i = 0; i < n_ch - 1; i++)
    draw_line_from_double_raw(im, &ch[i], &ch[i+1], COLOR_YELLOW);
  draw_line_from_double_raw(im, &ch[n_ch-1], &ch[0], COLOR_YELLOW);

  for(i = 0; i < n_ch; i++)
    draw_big_point_from_double_raw(im, &ch[i], COLOR_CYAN);
  
  return(1);
}

/*****************************************************************************/
