/*****************************************************************************/
/*                                                                           */
/* Author: Michael C. Nechyba                                                */
/* Last modified: 2/25/2001                                                  */
/*                                                                           */
/* File: "raw.c"                                                             */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef INCLUDE_JAVA
#include "GatorCameraClient.h"
#endif

#include "acticast_include.h"

/****************************** Subroutines **********************************/

int ppm2raw(PPM *ppm, RAW *raw, int raw_alloc)
{
  int i, j, size;
  char *buf;

  if(raw_alloc == 0) {
    // fprintf(stderr, "Allocating %d x %d\n", ppm->xdim, ppm->ydim);
    size = ppm->xdim*ppm->ydim*4;
    raw->p = (char *) malloc(sizeof(char)*size);
  }

  raw->xdim = ppm->xdim;
  raw->ydim = ppm->ydim;
  raw->xmin = ppm->xmin;
  raw->xmax = ppm->xmax;
  raw->ymin = ppm->ymin;
  raw->ymax = ppm->ymax;

  buf = raw->p;

  for(i = 0; i < ppm->ydim; i++) {
    for(j = 0; j < ppm->xdim; j++) {
      buf[3] = (char) 0;
      buf[2] = ppm->p[j][i][0];
      buf[1] = ppm->p[j][i][1];
      buf[0] = ppm->p[j][i][2];
      buf += 4;
    }
  }
  
  return(1);
}

/*****************************************************************************/

int raw2ppm(RAW *raw, PPM *ppm)
{
  int i, j;
  char *buf;

  ppm->xdim = raw->xdim;
  ppm->ydim = raw->ydim;
  ppm->xmin = raw->xmin;
  ppm->xmax = raw->xmax;
  ppm->ymin = raw->ymin;
  ppm->ymax = raw->ymax;

  buf = raw->p;

  for(i = 0; i < raw->ydim; i++) {
    for(j = 0; j < raw->xdim; j++) {
      ppm->p[j][i][0] = buf[2];
      ppm->p[j][i][1] = buf[1];
      ppm->p[j][i][2] = buf[0];
      buf += 4;
    }
  }
  
  return(1);
}

/*****************************************************************************/

int raw2ppm_small(RAW *raw, PPM *ppm)
{
  int i, j, loc;
  
  ppm->xdim = raw->xdim;
  ppm->ydim = raw->ydim;
  ppm->xmin = raw->xmin;
  ppm->xmax = raw->xmax;
  ppm->ymin = raw->ymin;
  ppm->ymax = raw->ymax;

  for(i = raw->ymin; i < raw->ymax; i++) {
    for(j = raw->xmin; j < raw->xmax; j++) {
      loc = (i*raw->xdim + j)*4;
      ppm->p[j][i][0] = raw->p[loc+2];
      ppm->p[j][i][1] = raw->p[loc+1];
      ppm->p[j][i][2] = raw->p[loc];
    }
  }
  
  return(1);
}

/*****************************************************************************/

int write_raw_file(char *file, RAW *raw)
{
  char *buf;
  int i, size;
  FILE *fp;
	
  fp = fopen(file, "w");
  if(fp == NULL) {
    fprintf(stderr, "Could not open %s for writing.\n", file);
    exit(1);
  }		

  size = raw->xdim*raw->ydim*4;
  buf = raw->p;

  for(i = 0; i < size; i++) {
    fprintf(fp, "%c", buf[i]);
  }

  fclose(fp);

  return(1);	
}

/*****************************************************************************/

int read_raw_file(char *file, RAW *raw, int xdim, int ydim, int raw_alloc)
{
  char *buf;
  int i, size;
  FILE *fp;
  
  raw->xdim = xdim;
  raw->ydim = ydim;
  
  size = raw->xdim*raw->ydim*4;

  if(raw_alloc == 0) {
    raw->p = (char *) malloc(sizeof(char)*size);
  }

  fp = fopen(file, "r");
  
  if(fp == NULL) {
    fprintf(stderr, "Could not read %s file.\n", file);
    exit(-1);
  }
  
  buf = raw->p;

  for(i = 0; i < size; i++) {
    fscanf(fp, "%c", &buf[i]);
  }
  
  fclose(fp);
  
  return(1);	
}

/*****************************************************************************/

int difference_raw(RAW *im2, RAW *im1, PPM *imout)
{
  int i, j, loc;
  char *buf1, *buf2;
  
  limit_dimensions_raw(im1);
  copy_dimensions_raw_to_ppm(im1, imout);
  
  for(i = im1->ymin; i < im1->ymax; i++) {
    for(j = im1->xmin; j < im1->xmax; j++) {
      loc = (i*im1->xdim + j)*4;
      buf1 = &im1->p[loc];
      buf2 = &im2->p[loc];
      
      imout->p[j][i][0] = difference_one_third_pixel(buf2[2], buf1[2]);
      imout->p[j][i][1] = difference_one_third_pixel(buf2[1], buf1[1]);
      imout->p[j][i][2] = difference_one_third_pixel(buf2[0], buf1[0]);
    }
  }

  return(1);
}

/*****************************************************************************/

int limit_dimensions_raw(RAW *im)
{
  if(im->xmin < 0)
    im->xmin = 0;
  
  if(im->xmax > im->xdim-1)
    im->xmax = im->xdim-1;
  
  if(im->ymin < 0)
    im->ymin = 0;
  
  if(im->ymax > im->ydim-1)
    im->ymax = im->ydim-1;
  
  return(1);
}

/*****************************************************************************/

int copy_dimensions_raw_to_ppm(RAW *imin, PPM *imout)
{
  imout->xdim = imin->xdim;
  imout->ydim = imin->ydim;
  imout->xmin = imin->xmin;
  imout->xmax = imin->xmax;
  imout->ymin = imin->ymin;
  imout->ymax = imin->ymax;
  
  return(1);
}

/*****************************************************************************/

int copy_dimensions_raw_to_raw(RAW *imin, RAW *imout)
{
  imout->xdim = imin->xdim;
  imout->ydim = imin->ydim;
  imout->xmin = imin->xmin;
  imout->xmax = imin->xmax;
  imout->ymin = imin->ymin;
  imout->ymax = imin->ymax;
  
  return(1);
}

/*****************************************************************************/

int set_dimensions_raw(RAW *im, int xmin, int xmax, int ymin, int ymax)
{
	im->xmin = xmin;
	im->xmax = xmax;
	im->ymin = ymin;
	im->ymax = ymax;

	return(1);
}

/*****************************************************************************/

int display_dim_raw(FILE *fp, RAW *im)
{
  fprintf(fp, "DIM DISPLAY: (%d, %d) --> (%d, %d); [0, %d] [0, %d]\n", 
	  im->xmin, im->ymin,
	  im->xmax, im->ymax, im->xdim, im->ydim);
	  
  return(1);  
}

/*****************************************************************************/

int filter_raw(RAW *imin, PPM *imout, FILTER *f)
{
  int i, j, k, l, fsize, tmp, loc;
  double r, g, b, w;	
  char *buf;

  fsize = f->size;
  
  limit_dimensions_raw(imin);
  copy_dimensions_raw_to_ppm(imin, imout);
  
  for(i = imin->ymin; i < imin->ymax; i++) {
    for(j = imin->xmin; j < imin->xmax; j++) {
      r = 0.;
      g = 0.;
      b = 0.;
      w = 0.;
      for(k = -fsize; k <= fsize; k++) {
	for(l = -fsize; l <= fsize; l++) {
	  if((i+k) >= imin->ymin && (i+k) < imin->ymax && 
	     (j+l) >= imin->xmin && (j+l) < imin->xmax) {
	    loc = ((i+k)*imin->xdim + (j+l))*4;
	    buf = &imin->p[loc];
	    tmp = convert_char(buf[2]);
	    r += (f->m[l+fsize][k+fsize]*((double) tmp));
	    
	    tmp = convert_char(buf[1]);
	    g += (f->m[l+fsize][k+fsize]*((double) tmp));
	    
	    tmp = convert_char(buf[0]);
	    b += (f->m[l+fsize][k+fsize]*((double) tmp));
	    
	    w += f->m[l+fsize][k+fsize];
	  }
	}
      }
      
      imout->p[j][i][0] = (int) (r/w + .5);
      imout->p[j][i][1] = (int) (g/w + .5);
      imout->p[j][i][2] = (int) (b/w + .5);
    }
  }
  
  return(1);
}

/*****************************************************************************/

int draw_line_raw(char *im, int xmin, int xmax, int ymin, int ymax,
		  int r, int g, int b)
{
  int x, y, loc, xmin_local, xmax_local, ymin_local, ymax_local;
  double dx, dy, m;
  
  dx = ((double) (xmax - xmin));
  dy = ((double) (ymax - ymin));
  
  if(dx > dy) {
    m = dy/dx;
    
    xmin_local = max_int(xmin, 0);
    xmax_local = min_int(xmax, X_DIM);

    for(x = xmin_local; x < xmax_local; x++) {
      y = (int) (((double) ymin) + m*((double) (x - xmin)) + .5);

      if(y >= 0 && y < Y_DIM) {
	loc = (y* X_DIM + x)*4;
	
	im[loc] = (char) b;
	im[loc+1] = (char) g;
	im[loc+2] = (char) r;
      }
    }
  } else {
    m = dx/dy;
    
    ymin_local = max_int(ymin, 0);
    ymax_local = min_int(ymax, Y_DIM);

    for(y = ymin_local; y < ymax_local; y++) {
      x = (int) (((double) xmin) + m*((double) (y - ymin)) + .5);
      if(x >= 0 && x < X_DIM) {
	loc = (y* X_DIM + x)*4;
      
	im[loc] = (char) b;
	im[loc+1] = (char) g;
	im[loc+2] = (char) r;
      }
    }
  }
  
  return(1);
}

/*****************************************************************************/

int draw_point_raw(char *im, POINT *p, int r, int g, int b)
{
  int x, y, loc;
  
  x = p->x;
  y = p->y;
  loc = (y* X_DIM + x)*4;
  
  if(x >= 0 && x < X_DIM  && y >= 0 && y < Y_DIM) {
    im[loc] = (char) b;
    im[loc+1] = (char) g;
    im[loc+2] = (char) r;
  }
  return(1);
}

/*****************************************************************************/

int draw_big_point_raw(char *im, POINT *p, int r, int g, int b)
{
  int x, y, loc, loc2;
  
  x = p->x;
  y = p->y;
  loc = (y* X_DIM + x)*4;

  if(x >= 0 && x < X_DIM && y >= 0 && y < Y_DIM) {
    im[loc] = (char) b;
    im[loc+1] = (char) g;
    im[loc+2] = (char) r;
  } else {
    return(-1);
  }
  
  if(x >= 1) {
    loc2 = (y* X_DIM + x-1)*4;
    im[loc2] = (char) b;
    im[loc2+1] = (char) g;
    im[loc2+2] = (char) r;
  }
  
  if(x < (X_DIM - 1)) {
    loc2 = (y* X_DIM + x+1)*4;
    im[loc2] = (char) b;
    im[loc2+1] = (char) g;
    im[loc2+2] = (char) r;
  }
  
  if(y >= 1) {
    loc2 = ((y-1)* X_DIM + x)*4;
    im[loc2] = (char) b;
    im[loc2+1] = (char) g;
    im[loc2+2] = (char) r;
  }
  
  if(y < (Y_DIM)) {
    loc2 = ((y+1)* X_DIM + x)*4;
    im[loc2] = (char) b;
    im[loc2+1] = (char) g;
    im[loc2+2] = (char) r;
  }
  
  return(1);
}

/*****************************************************************************/

int draw_region_box_raw(char *im, int r, int g, int b, BOUNDING_BOX *box)
{
  int xmin, xmax, ymin, ymax, loc;
  
  xmin = box->xmin;
  xmax = box->xmax;
  ymin = box->ymin;
  ymax = box->ymax;
  
  draw_line_raw(im, xmin, xmin, ymin, ymax, r, g, b);
  draw_line_raw(im, xmax, xmax, ymin, ymax, r, g, b);
  draw_line_raw(im, xmin, xmax, ymin, ymin, r, g, b);
  draw_line_raw(im, xmin, xmax, ymax, ymax, r, g, b);
  
  return(1);
}

/*****************************************************************************/

int draw_ellipse_raw(char *im, ELLIPSE *ell, int r, int g, int b)
{
  POINT p;
  double angle, dangle;
  double xtmp, ytmp, ct, st;

  dangle = 0.01;

  ct = cos(ell->theta);
  st = sin(ell->theta);

  for(angle = 0.; angle < 2.*M_PI; angle+=dangle) {
    xtmp = ell->a*cos(angle);
    ytmp = ell->b*sin(angle);

    p.x = (int) (ct*xtmp - st*ytmp + ell->mean[0] + 0.5);
    p.y = (int) (st*xtmp + ct*ytmp + ell->mean[1] + 0.5);

    p.x = limit_int(p.x, 0, X_DIM-1);
    p.y = limit_int(p.y, 0, Y_DIM-1);

    draw_big_point_raw(im, &p, r, g, b);
  }

  return(1);
}

/*****************************************************************************/

int draw_mask_raw(char *im, int r, int g, int b, BOUNDING_BOX *box, PPM *mask)
{
  int x, y, loc;

  for (x = box->xmin; x < box->xmax; x++)
    {
      for (y = box->ymin; y < box->ymax; y++)
        {
          loc = (y* X_DIM + x)*4;
          if (mask->p[x][y][0]){
            im[loc] = b;
            im[loc+1] = g;
            im[loc+2] = r;
          }
        }
    }
}

/*****************************************************************************/
